/****************************************
Daggorath PC-Port Version 0.2.1
Richard Hunerlach
November 13, 2002

The copyright for Dungeons of Daggorath
is held by Douglas J. Morgan.
(c) 1982, DynaMicro
*****************************************/

// Dungeons of Daggorath
// PC-Port
// Filename: creature.cpp
//
// Implementation of Creature class

#include "creature.h"
#include "dodgame.h"
#include "dungeon.h"
#include "object.h"
#include "sched.h"
#include "viewer.h"
#include "rng.h"
#include "player.h"
#include "oslink.h"

extern OS_Link		oslink;
extern Player		player;
extern RNG			rng;
extern Viewer		viewer;
extern dodGame		game;
extern Dungeon		dungeon;
extern Object		object;
extern Scheduler	scheduler;

// Constructor
Creature::Creature() : CTYPES(12),
					   CMXPTR(0),
					   CC_LEN(18),
					   FRZFLG(0),
					   creChannel(1),
					   creChannelv(2),
					   creSpeedMul(100)
{
	CDBTAB[0] = CDB(32,0,255,128,255,2300,1100);
	CDBTAB[1] = CDB(56,0,255,80,128,1500,700);
	CDBTAB[2] = CDB(200,0,255,52,192,2900,2300);
	CDBTAB[3] = CDB(304,0,255,96,167,3100,3100);
	CDBTAB[4] = CDB(504,0,128,96,60,1300,700);
	CDBTAB[5] = CDB(704,0,128,128,48,1700,1300);
	CDBTAB[6] = CDB(400,255,128,255,128,500,400);
	CDBTAB[7] = CDB(800,0,64,255,8,1300,700);
	CDBTAB[8] = CDB(800,192,16,192,8,300,300);
	CDBTAB[9] = CDB(1000,255,5,255,3,400,300);
	CDBTAB[10] = CDB(1000,255,6,255,0,1300,700);
	CDBTAB[11] = CDB(8000,255,6,255,0,1300,700);

	// Should move this to Utils
	char vals[] = "994200000000240666000000000406840010000000866400222222244801";
	int v = 0;
	while (vals[v]) { CMXLND[v] = vals[v++] - '0'; }

	strcpy(vals,"0310130");
	v = 0;
	while (vals[v]) { MOVTAB[v] = vals[v++] - '0'; }
}

void Creature::Reset()
{
	CMXPTR = 0;
	FRZFLG = 0;
	
	char vals[] = "994200000000240666000000000406840010000000866400222222244801";
	int v = 0;
	while (vals[v]) { CMXLND[v] = vals[v++] - '0'; }

	strcpy(vals,"0310130");
	v = 0;
	while (vals[v]) { MOVTAB[v] = vals[v++] - '0'; }
}

void Creature::UpdateCreSpeed()
{
	for (int ccc = 0; ccc < 12; ++ccc)
	{
		CDBTAB[ccc].P_CDTMV = ((int)((float)CDBTAB[ccc].P_CDTMV*(float)creSpeedMul/100.0));
		CDBTAB[ccc].P_CDTAT = ((int)((float)CDBTAB[ccc].P_CDTAT*(float)creSpeedMul/100.0));
	}
}

void Creature::LoadSounds()
{
	char fn[20];
	char fpre[7];
	strcpy(fpre,oslink.soundDir);
	strcat(fpre,oslink.pathSep);

	strcpy(fn,fpre);
	strcat(fn,"00_squeak.wav");
	creSound[0] = Mix_LoadWAV(fn);

	strcpy(fn,fpre);
	strcat(fn,"01_rattle.wav");
	creSound[1] = Mix_LoadWAV(fn);
	
	strcpy(fn,fpre);
	strcat(fn,"02_growl.wav");
	creSound[2] = Mix_LoadWAV(fn);
	
	strcpy(fn,fpre);
	strcat(fn,"03_beoop.wav");
	creSound[3] = Mix_LoadWAV(fn);
	
	strcpy(fn,fpre);
	strcat(fn,"04_klank.wav");
	creSound[4] = Mix_LoadWAV(fn);
	
	strcpy(fn,fpre);
	strcat(fn,"05_grawl.wav");
	creSound[5] = Mix_LoadWAV(fn);
	
	strcpy(fn,fpre);
	strcat(fn,"06_pssst.wav");
	creSound[6] = Mix_LoadWAV(fn);
	
	strcpy(fn,fpre);
	strcat(fn,"07_kklank.wav");
	creSound[7] = Mix_LoadWAV(fn);
	
	strcpy(fn,fpre);
	strcat(fn,"08_pssht.wav");
	creSound[8] = Mix_LoadWAV(fn);
	
	strcpy(fn,fpre);
	strcat(fn,"09_snarl.wav");
	creSound[9] = Mix_LoadWAV(fn);
	
	strcpy(fn,fpre);
	strcat(fn,"0A_bdlbdl.wav");
	creSound[10] = Mix_LoadWAV(fn);
	
	strcpy(fn,fpre);
	strcat(fn,"0B_bdlbdl.wav");
	creSound[11] = Mix_LoadWAV(fn);

	strcpy(fn,fpre);
	strcat(fn,"13_clank.wav");
	clank = Mix_LoadWAV(fn);

	strcpy(fn,fpre);
	strcat(fn,"16_kaboom.wav");
	kaboom = Mix_LoadWAV(fn);

	strcpy(fn,fpre);
	strcat(fn,"19_buzz.wav");
	buzz = Mix_LoadWAV(fn);
}

// This routine creates a new dungeon level,
// filling it with objects and creatures.  It
// should probably be moved to the Dungeon class.
void Creature::NEWLVL()
{
	dodBYTE	a, b;
	int		u, idx, tmp;

	CMXPTR = game.LEVEL * CTYPES;
	dungeon.CalcVFI();
	for (tmp = 0; tmp < 32; ++tmp)
	{
		CCBLND[tmp].clear();
	}
	scheduler.SYSTCB();
	dungeon.DGNGEN();
	u = CMXPTR;
	a = CTYPES - 1;
	do
	{
		b = CMXLND[u + a];
		if (b != 0)
		{
			do
			{
				CBIRTH(a);
				--b;
			} while (b != 0);
		}
		--a;
	} while (a != 0xFF);

	u = -1;
	object.OFINDF = 0;
	do
	{
		idx = object.FNDOBJ();
		if (idx == -1)
		{
			break;
		}
		if (object.OCBLND[idx].P_OCOWN == 0xFF)
		{
			do
			{
				++u;
				if (u == 32)
				{
					u = 0;
				}
				if (CCBLND[u].P_CCUSE != 0)
				{
					tmp = CCBLND[u].P_CCOBJ;
					CCBLND[u].P_CCOBJ = idx;
					object.OCBLND[idx].P_OCPTR = tmp;
					break;
				}
			} while (true);
		}
	} while (true);

	// Determine VDGINV Setting
	if (game.LEVEL % 2 == 0)
	{
		viewer.VDGINV = false;
	}
	else
	{
		viewer.VDGINV = true;
	}
	viewer.setVid();
}

// Creates a new creature and places it in the maze
void Creature::CBIRTH(dodBYTE typ)
{
	int			u, maz_idx;
	RowCol		rndcell;
	dodBYTE		rw, cl;
	int			TCBindex;

	u = -1;
	do
	{
		++u;
	} while (CCBLND[u].P_CCUSE != 0);
	--CCBLND[u].P_CCUSE;

	CCBLND[u].P_CCTYP = typ;
	CCBLND[u].P_CCPOW = CDBTAB[typ].P_CDPOW;
	CCBLND[u].P_CCMGO = CDBTAB[typ].P_CDMGO;
	CCBLND[u].P_CCMGD = CDBTAB[typ].P_CDMGD;
	CCBLND[u].P_CCPHO = CDBTAB[typ].P_CDPHO;
	CCBLND[u].P_CCPHD = CDBTAB[typ].P_CDPHD;
	CCBLND[u].P_CCTMV = CDBTAB[typ].P_CDTMV;
	CCBLND[u].P_CCTAT = CDBTAB[typ].P_CDTAT;

	do
	{
		do
		{
			cl = (rng.RANDOM() & 31);
			rw = (rng.RANDOM() & 31);
			rndcell.setRC(rw, cl);
			maz_idx = dungeon.RC2IDX(rw, cl);
		} while (dungeon.MAZLND[maz_idx] == 0xFF);
	} while (CFIND(rw, cl) == false);
	
	CCBLND[u].P_CCROW = rw;
	CCBLND[u].P_CCCOL = cl;

	TCBindex = scheduler.GETTCB();
	scheduler.TCBLND[TCBindex].data = u;
	scheduler.TCBLND[TCBindex].type = scheduler.cmoveTCB;
	scheduler.TCBLND[TCBindex].frequency = CCBLND[u].P_CCTMV;
}

// These two routines should probably be combined.
// They check for a creature in the given cell
bool Creature::CFIND(dodBYTE rw, dodBYTE cl)
{
	int ctr = 0;
	while (ctr < 32)
	{
		if (CCBLND[ctr].P_CCROW == rw &&
			CCBLND[ctr].P_CCCOL == cl)
		{
			if (CCBLND[ctr].P_CCUSE != 0)
				return false;
		}
		++ctr;
	}
	return true;
}

// These two routines should probably be combined.
// They check for a creature in the given cell
int Creature::CFIND2(RowCol rc)
{
	int ctr = 0;
	while (ctr < 32)
	{
		if (CCBLND[ctr].P_CCROW == rc.row &&
			CCBLND[ctr].P_CCCOL == rc.col)
		{
			if (CCBLND[ctr].P_CCUSE != 0)
			{
				return ctr;
			}
		}
		++ctr;
	}
	return -1;
}

// This method is called from the scheduler once every five
// minutes.  It will generate random new creatures.
int Creature::CREGEN()
{
	// Update Task's next_time
	scheduler.TCBLND[scheduler.cregenTCB].next_time = 
		scheduler.curTime +
		scheduler.TCBLND[scheduler.cregenTCB].frequency;

	int X = CMXPTR;
	dodBYTE B = CTYPES - 1;
	dodBYTE A = 0;
	do
	{
		A += CMXLND[X + B];
		--B;
	} while (B != 255);
	if (A < 32)
	{
		A = rng.RANDOM();
		A &= 7;
		A += 2;
		++CMXLND[X + A];
	}
	return 0;
}

// This method is called from the scheduler to move the
// creatures.  This is where most of the creature logic
// resides.  It's frequency is determined by the creature
// type, and its location relative to the player.
int Creature::CMOVE(int task, int cidx)
{
	int oidx, dir, X, loop;
	bool doRandom = false;
	dodBYTE r, c, rnd, d;
	dodBYTE shA, shB;
	dodSHORT shD, shD2;

	if (FRZFLG == 0)
	{
		// ignore dead creatures
		if (CCBLND[cidx].P_CCUSE == 0)
		{
			return 0;
		}

		// pick up object
		if (CCBLND[cidx].P_CCTYP != 6 &&
			CCBLND[cidx].P_CCTYP < 10)
		{
			object.OFINDF = 0;
			oidx = object.OFIND(RowCol(CCBLND[cidx].P_CCROW,
									   CCBLND[cidx].P_CCCOL));
			if (oidx != -1)
			{
				object.OCBLND[oidx].P_OCPTR = CCBLND[cidx].P_CCOBJ;
				CCBLND[cidx].P_CCOBJ = oidx;
				--object.OCBLND[oidx].P_OCOWN;
				viewer.PUPDAT();
				if (CCBLND[cidx].P_CCROW == player.PROW &&
					CCBLND[cidx].P_CCCOL == player.PCOL)
				{
					viewer.PUPDAT();
					viewer.NEWLUK = 0;
					scheduler.TCBLND[task].next_time = scheduler.curTime +
						CCBLND[cidx].P_CCTAT;
					return 0;
				}
				else
				{
					scheduler.TCBLND[task].next_time = scheduler.curTime +
						CCBLND[cidx].P_CCTMV;
					return 0;
				}
			}
		}

		// attack player
		if (CCBLND[cidx].P_CCROW == player.PROW &&
			CCBLND[cidx].P_CCCOL == player.PCOL)
		{
			// do creature sound
			Mix_PlayChannel(creChannel, creSound[CCBLND[cidx].P_CCTYP], 0);
			while (Mix_Playing(creChannel) == 1)
			{
				if (scheduler.curTime >= scheduler.TCBLND[0].next_time)
				{
					scheduler.CLOCK();
				}
				scheduler.curTime = SDL_GetTicks();
			}

			// set player shielding parameters
			shA = 0x80;
			shB = 0x80;

			if (player.PLHAND != -1 && object.OCBLND[player.PLHAND].P_OCCLS == 3)
			{
				shD = ((shA << 8) | shB);
				shD2 = ((object.OCBLND[player.PLHAND].P_OCXX0 << 8) |
						 object.OCBLND[player.PLHAND].P_OCXX1);
				if (shD2 < shD)
				{
					shA = (shD2 >> 8);
					shB = (shD2 & 255);
				}
			}

			if (player.PRHAND != -1 && object.OCBLND[player.PRHAND].P_OCCLS == 3)
			{
				shD = ((shA << 8) | shB);
				shD2 = ((object.OCBLND[player.PRHAND].P_OCXX0 << 8) |
						 object.OCBLND[player.PRHAND].P_OCXX1);
				if (shD2 < shD)
				{
					shA = (shD2 >> 8);
					shB = (shD2 & 255);
				}
			}

			player.PMGD = shA;
			player.PPHD = shB;

			// process attack

			if (player.ATTACK(CCBLND[cidx].P_CCPOW, player.PPOW,
				player.PDAM))
			{
				// make CLANK sound
				Mix_PlayChannel(creChannel, clank, 0);
				while (Mix_Playing(creChannel) == 1)
				{
					if (scheduler.curTime >= scheduler.TCBLND[0].next_time)
					{
						scheduler.CLOCK();
					}
					scheduler.curTime = SDL_GetTicks();
				}

				player.DAMAGE(CCBLND[cidx].P_CCPOW, CCBLND[cidx].P_CCMGO,
					   CCBLND[cidx].P_CCPHO, player.PPOW,
					   player.PMGD, player.PPHD, &player.PDAM);
			}

			player.HUPDAT();
			scheduler.TCBLND[task].next_time = scheduler.curTime +
				CCBLND[cidx].P_CCTMV;
			return 0;
		}

		// look for player along ROW axis
		if (CCBLND[cidx].P_CCROW == player.PROW)
		{
			if (CCBLND[cidx].P_CCCOL < player.PCOL)
			{
				dir = 1;
			}
			else
			{
				dir = 3;
			}
			r = CCBLND[cidx].P_CCROW;
			c = CCBLND[cidx].P_CCCOL;
			do
			{
				if (!dungeon.STEPOK(r, c, dir))
				{
					doRandom = true;
					break;
				}
				r += dungeon.STPTAB[dir * 2];
				c += dungeon.STPTAB[(dir * 2) + 1];	
			} while (!(r == player.PROW &&
					   c == player.PCOL));
			if (doRandom == false)
			{
				CCBLND[cidx].P_CCDIR = dir;
				CWALK(0, &CCBLND[cidx]);
				if (CCBLND[cidx].P_CCROW == player.PROW &&
					CCBLND[cidx].P_CCCOL == player.PCOL)
				{
					viewer.PUPDAT();
					viewer.NEWLUK = 0;
					scheduler.TCBLND[task].next_time = scheduler.curTime +
						CCBLND[cidx].P_CCTAT;
					return 0;
				}
				else
				{
					scheduler.TCBLND[task].next_time = scheduler.curTime +
						CCBLND[cidx].P_CCTMV;
					return 0;
				}
			}
		}
		
		// look for player along COL axis
		if (CCBLND[cidx].P_CCCOL == player.PCOL)
		{
			if (CCBLND[cidx].P_CCROW < player.PROW)
			{
				dir = 2;
			}
			else
			{
				dir = 0;
			}
			r = CCBLND[cidx].P_CCROW;
			c = CCBLND[cidx].P_CCCOL;
			do
			{
				if (!dungeon.STEPOK(r, c, dir))
				{
					doRandom = true;
					break;
				}
				r += dungeon.STPTAB[dir * 2];
				c += dungeon.STPTAB[(dir * 2) + 1];	
			} while (!(r == player.PROW &&
					   c == player.PCOL));
			if (doRandom == false)
			{
				CCBLND[cidx].P_CCDIR = dir;
				CWALK(0, &CCBLND[cidx]);
				if (CCBLND[cidx].P_CCROW == player.PROW &&
					CCBLND[cidx].P_CCCOL == player.PCOL)
				{
					viewer.PUPDAT();
					viewer.NEWLUK = 0;
					scheduler.TCBLND[task].next_time = scheduler.curTime +
						CCBLND[cidx].P_CCTAT;
					return 0;
				}
				else
				{
					scheduler.TCBLND[task].next_time = scheduler.curTime +
						CCBLND[cidx].P_CCTMV;
					return 0;
				}
			}
		}

		// player not seen so make random move
		if (doRandom ||
			(CCBLND[cidx].P_CCROW != player.PROW &&
			 CCBLND[cidx].P_CCCOL != player.PCOL))
		{
			X = 0;
			rnd = rng.RANDOM();
			if ((rnd & 128) == 0)
			{
				X += 3;
			}
			rnd &= 3;
			if (rnd == 0)
			{
				X += 1;
			}
			loop = 3;
			do
			{
				d = MOVTAB[X++];
				if (CWALK(d, &CCBLND[cidx]))
				{
					if (CCBLND[cidx].P_CCROW == player.PROW &&
						CCBLND[cidx].P_CCCOL == player.PCOL)
					{
						viewer.PUPDAT();
						viewer.NEWLUK = 0;
						scheduler.TCBLND[task].next_time = scheduler.curTime +
							CCBLND[cidx].P_CCTAT;
						return 0;
					}
					else
					{
						scheduler.TCBLND[task].next_time = scheduler.curTime +
							CCBLND[cidx].P_CCTMV;
						return 0;
					}
				}
				--loop;
			} while (loop != 0);
			CWALK(2, &CCBLND[cidx]);
		}
	}

	if (CCBLND[cidx].P_CCROW == player.PROW &&
		CCBLND[cidx].P_CCCOL == player.PCOL)
	{
		viewer.PUPDAT();
		viewer.NEWLUK = 0;
		scheduler.TCBLND[task].next_time = scheduler.curTime +
			CCBLND[cidx].P_CCTAT;
		return 0;
	}
	else
	{
		scheduler.TCBLND[task].next_time = scheduler.curTime +
			CCBLND[cidx].P_CCTMV;
		return 0;
	}
}

// This routine attempts to move the creature in the
// given direction.
bool Creature::CWALK(dodBYTE dir, CCB * cr)
{
	dodBYTE DIR, r, c, rr, cc, big, small;

	dir += cr->P_CCDIR;
	dir &= 3;
	DIR = dir;

	r = cr->P_CCROW;
	c = cr->P_CCCOL;
	if (dungeon.STEPOK(r, c, DIR))
	{
		r += dungeon.STPTAB[DIR * 2];
		c += dungeon.STPTAB[(DIR * 2) + 1];
		if (!CFIND(r,c))
		{
			return false;
		}

		rr = r;
		cc = c;

		if (r > player.PROW)
		{
			r = r - player.PROW;
		}
		else
		{
			r = player.PROW - r;
		}

		if (c > player.PCOL)
		{
			c = c - player.PCOL;
		}
		else
		{
			c = player.PCOL - c;
		}

		if (r > c)
		{
			big = r;
			small = c;
		}
		else
		{
			big = c;
			small = r;
		}

		if (big > 8)
		{
			cr->P_CCROW = rr;
			cr->P_CCCOL = cc;
			cr->P_CCDIR = DIR;
			return true;
		}

		if (small > 2)
		{
			cr->P_CCROW = rr;
			cr->P_CCCOL = cc;
			cr->P_CCDIR = DIR;
			return true;
		}

		small = (rng.RANDOM() & 1);
		if (small == 1)
		{
			// make sound
			Mix_Volume(creChannelv, (MIX_MAX_VOLUME / 8) * (9 - big) );
			Mix_PlayChannel(creChannelv, creSound[cr->P_CCTYP], 0);
			while (Mix_Playing(creChannelv) == 1)
			{
				if (scheduler.curTime >= scheduler.TCBLND[0].next_time)
				{
					scheduler.CLOCK();
				}
				scheduler.curTime = SDL_GetTicks();
			}
		}

		cr->P_CCROW = rr;
		cr->P_CCCOL = cc;
		cr->P_CCDIR = DIR;

		--viewer.NEWLUK;
		return true;
	}

	return false;
}