/****************************************
Daggorath PC-Port Version 0.2.1
Richard Hunerlach
November 13, 2002

The copyright for Dungeons of Daggorath
is held by Douglas J. Morgan.
(c) 1982, DynaMicro
*****************************************/

// Dungeons of Daggorath
// PC-Port
// Filename: oslink.cpp
//
// Implementation of OS_Link class

#include <iostream>
#include <fstream>

using namespace std;

#include "oslink.h"
#include "dodgame.h"
#include "viewer.h"
#include "sched.h"
#include "player.h"
#include "dungeon.h"
#include "parser.h"
#include "object.h"
#include "creature.h"

extern Creature		creature;
extern Object		object;
extern Dungeon		dungeon;
extern Player		player;
extern Coordinate	crd;
extern Viewer		viewer;
extern dodGame		game;
extern Scheduler	scheduler;
extern Parser		parser;

// Constructor
OS_Link::OS_Link() : width(0), height(0), bpp(0), flags(0),
					 audio_rate(11025), audio_format(AUDIO_S8),
					 audio_channels(1), audio_buffers(512),
					 gamefileLen(50)
{
#ifdef MACOSX
	strcpy(pathSep,"/");
#else
	strcpy(pathSep,"\\");
#endif

	strcpy(confDir, "conf");
	strcpy(soundDir, "sound");
	strcpy(savedDir, "saved");
	memset(gamefile,0,gamefileLen);
}

// This routine will eventually need updated to allow
// user customization of screen size and resolution.
// It currently asks for an 800x600 screen size.
void OS_Link::init()
{
	ifstream	fin;
	int			in;
	char 		instr[64];
	char		fn[20];

	strcpy(fn,confDir);
	strcat(fn,pathSep);
	strcat(fn,"opts.txt");
	
	fin.open(fn);
	if (fin)
	{
		fin >> instr; sscanf(instr, "%d", &in); creature.creSpeedMul = in;
		fin >> instr; sscanf(instr, "%d", &in); player.turnDelay = in;
		fin >> instr; sscanf(instr, "%d", &in); player.moveDelay = in;
		fin.close();
		creature.UpdateCreSpeed();
	}

	Uint32 ticks1, ticks2;
	const SDL_VideoInfo * info = NULL;
	SDL_Surface * surface;
	if(SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER | SDL_INIT_AUDIO) < 0)
	{
		fprintf(stderr, "Video initialization failed: %s\n", SDL_GetError());
		quitSDL(1);
	}

	if(Mix_OpenAudio(audio_rate, audio_format, audio_channels, audio_buffers))
	{
		fprintf(stderr, "Unable to open audio!\n");
		quitSDL(1);
	}

	creature.LoadSounds();
	object.LoadSounds();
	scheduler.LoadSounds();
	player.LoadSounds();

	Mix_AllocateChannels(4);
	Mix_Volume(-1, MIX_MAX_VOLUME);
	
	info = SDL_GetVideoInfo();
	if(!info)
	{
		fprintf(stderr, "Video query failed: %s\n", SDL_GetError());
		quitSDL(1);
	}
	width = 800;
	height = (int) (width * 0.75);
	bpp = info->vfmt->BitsPerPixel;
	SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	flags = SDL_OPENGL | SDL_FULLSCREEN;
	if((surface = SDL_SetVideoMode(width, height, bpp, flags)) == 0)
	{
		fprintf(stderr, "Video mode set failed: %s\n", SDL_GetError());
		quitSDL(1);
	}
	crd.setCurWH((double) width);
	SDL_ShowCursor(SDL_DISABLE);
	viewer.setup_opengl();

	// Delay to wait for monitor to change modes if necessary
	// This ought to be made more intelligent
	ticks1 = SDL_GetTicks();
	do
	{
		ticks2 = SDL_GetTicks();
	} while (ticks2 < ticks1 + 2500);
	game.COMINI();
 	while (true)
	{
		scheduler.SCHED();
		if (scheduler.ZFLAG == 0xFF)
		{
			game.LoadGame();
			scheduler.ZFLAG = 0;
		}
		else
		{
			game.Restart();
		}
	}
}

// Used to check for keystrokes and application termination
void OS_Link::process_events()
{
	SDL_Event event;
	while(SDL_PollEvent(&event))
	{
		switch(event.type)
		{
		case SDL_KEYDOWN:
			handle_key_down(&event.key.keysym);
			break;
		case SDL_QUIT:
			quitSDL(0);
			break;
		}
	}
}

// Quits application
void OS_Link::quitSDL(int code)
{
	Mix_CloseAudio();
	SDL_Quit();
	exit(code);
}

// Processes key strokes.
void OS_Link::handle_key_down(SDL_keysym * keysym)
{
	dodBYTE c;
	if (viewer.DSPMOD == 0)
	{
		switch(keysym->sym)
		{
		case SDLK_ESCAPE:
			quitSDL(0);		// eventually change to meta-menu
			break;
		default:
			viewer.DSPMOD = 1;
			--viewer.UPDATE;
			parser.KBDPUT(32); // This is a (necessary ???) hack.
			break;
		}

	}
	else
	{
		switch(keysym->sym)
		{
		case SDLK_a: c = 'A'; break;
		case SDLK_b: c = 'B'; break;
		case SDLK_c: c = 'C'; break;
		case SDLK_d: c = 'D'; break;
		case SDLK_e: c = 'E'; break;
		case SDLK_f: c = 'F'; break;
		case SDLK_g: c = 'G'; break;
		case SDLK_h: c = 'H'; break;
		case SDLK_i: c = 'I'; break;
		case SDLK_j: c = 'J'; break;
		case SDLK_k: c = 'K'; break;
		case SDLK_l: c = 'L'; break;
		case SDLK_m: c = 'M'; break;
		case SDLK_n: c = 'N'; break;
		case SDLK_o: c = 'O'; break;
		case SDLK_p: c = 'P'; break;
		case SDLK_q: c = 'Q'; break;
		case SDLK_r: c = 'R'; break;
		case SDLK_s: c = 'S'; break;
		case SDLK_t: c = 'T'; break;
		case SDLK_u: c = 'U'; break;
		case SDLK_v: c = 'V'; break;
		case SDLK_w: c = 'W'; break;
		case SDLK_x: c = 'X'; break;
		case SDLK_y: c = 'Y'; break;
		case SDLK_z: c = 'Z'; break;

		case SDLK_BACKSPACE:
			c = parser.C_BS;
			break;
		case SDLK_RETURN:
			c = parser.C_CR;
			break;
		case SDLK_SPACE:
			c = parser.C_SP;
			break;
		case SDLK_ESCAPE:
			quitSDL(0);		// eventually change to meta-menu
			break;
		default:
			break;
		}
		parser.KBDPUT(c);
	}
}
