/****************************************
Daggorath PC-Port Version 0.2.1
Richard Hunerlach
November 13, 2002

The copyright for Dungeons of Daggorath
is held by Douglas J. Morgan.
(c) 1982, DynaMicro
*****************************************/

// Dungeons of Daggorath
// PC-Port
// Filename: player.cpp
//
// Implementation of Player class

#include "player.h"
#include "dodgame.h"
#include "viewer.h"
#include "sched.h"
#include "parser.h"
#include "object.h"
#include "dungeon.h"
#include "creature.h"
#include "oslink.h"

extern OS_Link		oslink;
extern Creature		creature;
extern Object		object;
extern Viewer		viewer;
extern dodGame		game;
extern Scheduler	scheduler;
extern Parser		parser;
extern Dungeon		dungeon;

// Constructor
Player::Player() : PROW(12), PCOL(22),
				   POBJWT(35), FAINT(0),
				   PPOW(PLRBLK.P_ATPOW),
				   PMGO(PLRBLK.P_ATMGO),
				   PMGD(PLRBLK.P_ATMGD),
				   PPHO(PLRBLK.P_ATPHO),
				   PPHD(PLRBLK.P_ATPHD),
				   PLHAND(PLRBLK.P_ATXX1),
				   PRHAND(PLRBLK.P_ATXX2),
				   PDAM(PLRBLK.P_ATDAM),
				   PRLITE(0), PMLITE(0),
				   turnDelay(20),
				   moveDelay(25),
				   wizDelay(500)
{
	PPOW = ((0x17 << 8) | 160);
	EMPHND.P_OCCLS = 4;
	EMPHND.P_OCREV = 0;
	EMPHND.P_OCMGO = 0;
	EMPHND.P_OCPHO = 5;
	PLHAND = -1;
	PRHAND = -1;
	PTORCH = -1;
	BAGPTR = -1;
}

void Player::Reset()
{
	POBJWT = 35;
	FAINT = 0;
	PRLITE = 0;
	PMLITE = 0;
	PLHAND = -1;
	PRHAND = -1;
	PTORCH = -1;
	BAGPTR = -1;
	EMPHND.P_OCCLS = 4;
	EMPHND.P_OCREV = 0;
	EMPHND.P_OCMGO = 0;
	EMPHND.P_OCPHO = 5;
	PDAM = 0;
	PDIR = 0;
	HEARTF = 0;
	HEARTC = 0;
	HEARTR = 0;
	HEARTS = 0;
	HBEATF = 0;
}

// Public Interface

void Player::LoadSounds()
{
	char fn[20];
	char fpre[7];
	strcpy(fpre,oslink.soundDir);
	strcat(fpre,oslink.pathSep);

	strcpy(fn,fpre);
	strcat(fn,"12_klink.wav");
	klink = Mix_LoadWAV(fn);

	strcpy(fn,fpre);
	strcat(fn,"14_thud.wav");
	thud = Mix_LoadWAV(fn);

	strcpy(fn,fpre);
	strcat(fn,"15_bang.wav");
	bang = Mix_LoadWAV(fn);
}

// This method gets called from the scheduler as often
// as possible.  It retrieves keyboard input, or commands
// from the demo data.
int Player::PLAYER()
{
	// Update Task's next_time
	scheduler.TCBLND[scheduler.playerTCB].next_time = scheduler.curTime +
		scheduler.TCBLND[scheduler.playerTCB].frequency;
	
	dodBYTE c;
	if (game.AUTFLG == 0)
	{
		// Process Keyboard Buffer
		do
		{
			c = parser.KBDGET();
			if (c == 0)
			{
				return 0;
			}
			if (FAINT != 0)
			{
				while (parser.KBDGET() != 0)
					;	// loop !!!
				return 0;
			}

			// Convert from ASCII to Internal Codes
			if (c >= 'A' && c <= 'Z')
			{
				c &= 0x1F;
			}
			else if (c == parser.C_BS)
			{
				c = parser.I_BS;
			}
			else if (c == parser.C_CR)
			{
				c = parser.I_CR;
			}
			else
			{
				c = parser.I_SP;
			}

			HUMAN(c);
		} while (true);
	}
	else
	{
		// Process Autoplay Commands


		// Feed next autoplay command to HUMAN

	}

	return 0;
}

void Player::HUMAN(dodBYTE c)
{
	int		res;
	dodBYTE	A, B;
	
	// Check if we are displaying the map
	if (HEARTF == 0)
	{
		game.INIVU();
		viewer.PROMPT();
	}
	if (c == parser.I_CR)
	{
carriage_return:
		viewer.OUTCHR(parser.I_SP);
		parser.LINBUF[parser.LINPTR] = 0xFF;
		parser.LINBUF[parser.LINPTR+1] = 0xFF;
		parser.LINPTR = 0;

		// dispatch to proper routine
		res = parser.PARSER(&parser.CMDTAB[0], &A, &B, true);
		if (res == 1)
		{
			// dispatch
			switch (A)
			{
			case 0:
				PATTK();
				break;
			case 1:
				PCLIMB();
				break;
			case 2:
				PDROP();
				break;
			case 3:
				PEXAM();
				break;
			case 4:
				PGET();
				break;
			case 5:
				PINCAN();
				break;
			case 6:
				PLOOK();
				break;
			case 7:
				PMOVE();
				break;
			case 8:
				PPULL();
				break;
			case 9:
				PREVEA();
				break;
			case 10:
				PSTOW();
				break;
			case 11:
				PTURN();
				break;
			case 12:
				PUSE();
				break;
			case 13:
				PZLOAD();
				break;
			case 14:
				PZSAVE();
				break;
			}
		}
		if (res == -1)
		{
			parser.CMDERR();
		}

		if (HEARTF != 0 & FAINT == 0)
		{
			viewer.PROMPT();
		}

		parser.LINPTR = 0;
		return;
	}
	if (c == parser.I_BS)
	{
		if (parser.LINPTR == 0)
		{
			return;
		}
		--parser.LINPTR;
		viewer.OUTSTR(parser.M_ERAS);
		return;
	}
	// Buffer normal characters
	viewer.OUTCHR(c);
	parser.LINBUF[parser.LINPTR] = c;
	++parser.LINPTR;
	viewer.OUTSTR(parser.M_CURS);
	if (parser.LINPTR >= 32)
	{
		goto carriage_return;
	}
}

// This method gets called from the scheduler with a
// frequency equal to the current heart rate.  It performs
// damage recovery, indicated by slowing the heartbeat.
int Player::HSLOW()
{
	PLRBLK.P_ATDAM -= (PLRBLK.P_ATDAM >> 6);
	if ((PLRBLK.P_ATDAM & 0x8000) != 0)
	{
		PLRBLK.P_ATDAM = 0;
	}
	HUPDAT();

	// Update Task's next_time
	scheduler.TCBLND[scheduler.hslowTCB].next_time = scheduler.curTime +
		(HEARTR * 17);
	
	return 0;
}

// Calculates heartbeat based on power level and damage.
// Also processes fainting and death.
void Player::HUPDAT()
{
	Uint32 ticks1;
	SDL_Event event;

	// Heartrate in source:
	//
	//               PPOW * 64
	// HEARTR = ------------------- - 19
	//           PPOW + (PDAM * 2)
	//
	// The original division routine added one to the integer quotient
	// so that [(1/5) == 1], [(5/5) == 2], [(10/5) == 3], etc.
	// The formula below reflects that peculiarity by only subtracting 18.
	//
	HEARTR = (((PLRBLK.P_ATPOW) * 64) /
			  ((PLRBLK.P_ATPOW) + ((PLRBLK.P_ATDAM) * 2))) - 18;

	if (FAINT == 0)
	{
		// not in a faint
		if (HEARTR <= 3 || (HEARTR & 128) != 0)
		{
			// do faint
			FAINT = -1;
			viewer.clearArea(&viewer.TXTPRI);
			viewer.OLIGHT = viewer.RLIGHT;
			do
			{
				--viewer.MLIGHT;
				--viewer.UPDATE;
				viewer.draw_game();
				--viewer.RLIGHT;
				ticks1 = SDL_GetTicks();
				scheduler.curTime = ticks1;
				do
				{
					if (scheduler.curTime >= scheduler.TCBLND[0].next_time)
					{
						scheduler.CLOCK();
						scheduler.EscCheck();
					}
					scheduler.curTime = SDL_GetTicks();
				} while (scheduler.curTime < ticks1 + 750);
			} while (viewer.RLIGHT != 248);	// not equal to -8
			--viewer.UPDATE;
			parser.KBDHDR = 0;
			parser.KBDTAL = 0;
		}
	}
	else
	{
		// in a faint
		if (HEARTR >= 4 && (HEARTR & 128) == 0)
		{
			// do recover from faint
			do
			{
				--viewer.UPDATE;
				viewer.draw_game();
				++viewer.MLIGHT;
				++viewer.RLIGHT;
				ticks1 = SDL_GetTicks();
				scheduler.curTime = ticks1;
				do
				{
					if (scheduler.curTime >= scheduler.TCBLND[0].next_time)
					{
						scheduler.CLOCK();
						scheduler.EscCheck();
					}
					scheduler.curTime = SDL_GetTicks();
				} while (scheduler.curTime < ticks1 + 750);
			} while (viewer.RLIGHT != viewer.OLIGHT);
			FAINT = 0;
			viewer.PROMPT();
			--viewer.UPDATE;
		}
	}
	if (PLRBLK.P_ATPOW < PLRBLK.P_ATDAM)
	{
		// Do death
		while(SDL_PollEvent(&event))
		{
			; // clear event buffer
		}
		viewer.clearArea(&viewer.TXTSTS);
		viewer.clearArea(&viewer.TXTPRI);
		scheduler.deathFadeLoop();
	}
}

// This method is called every five seconds.  It will
// manage the lit torch's timers.
//
// The full burn time is stored in XX0 in 5-second units,
// so 15 minutes equals 180 5-second units.
//
// The magical light values and physical light values are
// stored in minutes.  Hence the conversion half way thorugh.
//
// This is a hack to give the main timer more granularity,
// so that it doesn't lose a whole minute on each level change,
// but I had to leave the other values as minutes because they
// are used elsewhere in a complicated formula to determine
// the ligthing (line pixelation/fade) values for the 3D viewer.
//
int Player::BURNER()
{
	// Update Task's next_time
	scheduler.TCBLND[scheduler.burnerTCB].next_time = 
		scheduler.curTime +
		scheduler.TCBLND[scheduler.burnerTCB].frequency;

	dodSHORT A;
	if (PTORCH == -1)
	{
		--viewer.NEWLUK;
		return 0;
	}
	A = object.OCBLND[PTORCH].P_OCXX0;
	if (A == 0)
	{
		--viewer.NEWLUK;
		return 0;
	}
	--A;
	object.OCBLND[PTORCH].P_OCXX0 = A;

	// Convert A to minutes
	(A % 12 == 0) ? A /= 12 : A = (A / 12) + 1;

	if (A <= 5)
	{
		object.OCBLND[PTORCH].P_OCTYP = 0x18;
		object.OCBLND[PTORCH].P_OCREV = 0;
	}
	if (A < object.OCBLND[PTORCH].P_OCXX1)
	{
		object.OCBLND[PTORCH].P_OCXX1 = A;
	}
	if (A < object.OCBLND[PTORCH].P_OCXX2)
	{
		object.OCBLND[PTORCH].P_OCXX2 = A;
	}
	--viewer.NEWLUK;
	return 0;
}

// Used during initialization to set data for either
// the built-in demo or starting a live game
void Player::setInitialObjects(bool isDemo)
{
	int x, y;
	
	if (isDemo)
	{
		game.LEVEL = 3;
		// demo: iron sword, pine torch, leather shield

		x = object.OBIRTH(0x0D, 0);
		++object.OCBLND[x].P_OCOWN;
		object.OCBFIL(0x0D, x);
		object.OCBLND[x].P_OCREV = 0;
		BAGPTR = x;
		y = x;

		x = object.OBIRTH(0x0F, 0);
		++object.OCBLND[x].P_OCOWN;
		object.OCBFIL(0x0F, x);
		object.OCBLND[x].P_OCREV = 0;
		object.OCBLND[y].P_OCPTR = x;
		y = x;

		x = object.OBIRTH(0x10, 0);
		++object.OCBLND[x].P_OCOWN;
		object.OCBFIL(0x10, x);
		object.OCBLND[x].P_OCREV = 0;
		object.OCBLND[y].P_OCPTR = x;
	}
	else
	{
		game.LEVEL = 0;
		PROW = 0x10;
		PCOL = 0x0B;
		PLRBLK.P_ATPOW = 160;

		x = object.OBIRTH(0x11, 0);
		++object.OCBLND[x].P_OCOWN;
		object.OCBFIL(0x11, x);
		object.OCBLND[x].P_OCREV = 0;
		BAGPTR = x;
		y = x;

		x = object.OBIRTH(0x0F, 0);
		++object.OCBLND[x].P_OCOWN;
		object.OCBFIL(0x0F, x);
		object.OCBLND[x].P_OCREV = 0;
		object.OCBLND[y].P_OCPTR = x;
	}
}

// Processes ATTACK command
void Player::PATTK()
{
	int res, idx, cidx, optr, val;
	OCB *U, *Y;
	dodBYTE r,c;
	SDL_Event event;
	Uint32 ticks1, ticks2;

	res = parser.PARHND();
	if (res == -1)
	{
		return;
	}

	if (res == 0)
	{
		idx = PLHAND;
		if (idx == -1)
		{
			U = &EMPHND;
		}
		else
		{
			U = &object.OCBLND[idx];
		}
	}
	else
	{
		idx = PRHAND;
		if (idx == -1)
		{
			U = &EMPHND;
		}
		else
		{
			U = &object.OCBLND[idx];
		}
	}

	Y = U;
	PMGO = U->P_OCMGO;
	PPHO = U->P_OCPHO;
	PDAM += ((PPOW * (((int) PMGO + (int) PPHO) / 8)) >> 7);

	// make sound for appropriate object
	Mix_PlayChannel(object.objChannel,
		object.objSound[U->P_OCCLS], 0);
	while (Mix_Playing(object.objChannel) == 1)
	{
		if (scheduler.curTime >= scheduler.TCBLND[0].next_time)
		{
			scheduler.CLOCK();
		}
		scheduler.curTime = SDL_GetTicks();
	}

	if (U->P_OCTYP >= 19 && U->P_OCTYP <= 21)
	{
		--U->P_OCXX0;
		if (U->P_OCXX0 == 0)
		{
			U->P_OCTYP = 0x16;
			object.OCBFIL(0x16, idx);
			U->P_OCREV = 0;
			viewer.STATUS();
		}
	}

	cidx = creature.CFIND2(RowCol(PROW, PCOL));
	if (cidx == -1)
	{
		HUPDAT();
		return;
	}

	if (!ATTACK(PLRBLK.P_ATPOW, creature.CCBLND[cidx].P_CCPOW,
		creature.CCBLND[cidx].P_CCDAM))
	{
		HUPDAT();
		return;
	}

	if (PTORCH == -1 || object.OCBLND[PTORCH].P_OCTYP == 0x18)
	{
		if ((rng.RANDOM() & 3) != 0)
		{
			HUPDAT();
			return;
		}
	}

	// make KLINK sound
	Mix_PlayChannel(object.objChannel, klink, 0);
	while (Mix_Playing(object.objChannel) == 1)
	{
		if (scheduler.curTime >= scheduler.TCBLND[0].next_time)
		{
			scheduler.CLOCK();
		}
		scheduler.curTime = SDL_GetTicks();
	}

	viewer.OUTSTI(viewer.exps);

	// do damage
	if (DAMAGE(PLRBLK.P_ATPOW, PLRBLK.P_ATMGO, PLRBLK.P_ATPHO,
			   creature.CCBLND[cidx].P_CCPOW,
			   creature.CCBLND[cidx].P_CCMGD,
			   creature.CCBLND[cidx].P_CCPHD,
			   &creature.CCBLND[cidx].P_CCDAM) == true)
	{
		// Creature still alive
		HUPDAT();
		return;
	}

	optr = creature.CCBLND[cidx].P_CCOBJ;
	while (optr != -1)
	{
		object.OCBLND[optr].P_OCOWN = 0;
		object.OCBLND[optr].P_OCROW = creature.CCBLND[cidx].P_CCROW;
		object.OCBLND[optr].P_OCCOL = creature.CCBLND[cidx].P_CCCOL;
		optr = object.OCBLND[optr].P_OCPTR;
	}

	--creature.CMXLND[creature.CMXPTR + creature.CCBLND[cidx].P_CCTYP];
	creature.CCBLND[cidx].P_CCUSE = 0;
	viewer.PUPDAT();

	// do loud explosion sound
	Mix_PlayChannel(object.objChannel, bang, 0);
	while (Mix_Playing(object.objChannel) == 1)
	{
		if (scheduler.curTime >= scheduler.TCBLND[0].next_time)
		{
			scheduler.CLOCK();
		}
		scheduler.curTime = SDL_GetTicks();
	}

	PPOW += (creature.CCBLND[cidx].P_CCPOW >> 3);
	if ((PPOW & 0x8000) != 0)
	{
		PPOW = 0x7FFF;
	}

	if (creature.CCBLND[cidx].P_CCTYP == 10)
	{
		// Wizard's Image Killed
		// transport to 4th level

		// do fade in with message
		
		// Pause so player can see scroll
		ticks1 = SDL_GetTicks();
		do
		{
			ticks2 = SDL_GetTicks();
		} while (ticks2 < ticks1 + wizDelay);
		
		while(SDL_PollEvent(&event))
		{
			; // clear event buffer
		}
		viewer.clearArea(&viewer.TXTSTS);
		viewer.clearArea(&viewer.TXTPRI);
		viewer.done = false;
		viewer.fadeVal = -2;
		viewer.VCTFAD = 32;
		viewer.VXSCAL = 0x80;
		viewer.VYSCAL = 0x80;

		// Start buzz
		Mix_Volume(viewer.fadChannel, 0);
		Mix_PlayChannel(viewer.fadChannel, creature.buzz, -1);

		while (!viewer.done)
		{
			viewer.enough_fade();
			scheduler.EscCheck();
		}

		// Stop buzz
		Mix_HaltChannel(viewer.fadChannel);

		BAGPTR = PTORCH;
		if (PTORCH != -1)
		{
			object.OCBLND[PTORCH].P_OCPTR = -1;
		}

		POBJWT = 200;
		game.LEVEL = 3;
		creature.NEWLVL();

		do
		{
			c = (rng.RANDOM() & 31);
			r = (rng.RANDOM() & 31);
			val = dungeon.MAZLND[dungeon.RC2IDX(r, c)];
		} while (val == 0xFF);

		PROW = r;
		PCOL = c;

		game.INIVU();
	}

	if (creature.CCBLND[cidx].P_CCTYP != 11)
	{
		HUPDAT();
		return;
	}

	// Killed the real Wizard
	--creature.FRZFLG;
	PRLITE = 0x07;
	PMLITE = 0x13;
	object.OCBPTR = 1;
	BAGPTR = -1;
	PTORCH = -1;
	PRHAND = -1;
	PLHAND = -1;
	game.INIVU();
	HUPDAT();
	return;
}

// Determines if an attack strikes its target
bool Player::ATTACK(int AP, int DP, int DD)
{
	int pidx, adjust, ret;
	int T0 = 15;
	int Dval = (DP - DD) * 4;

	do
	{
		Dval -= AP;
		if (Dval < 0)
		{
			break;
		}
		--T0;
	} while (T0 > 0);

	pidx = T0 - 3;
	if (pidx >= 0)
	{
		adjust = T0 * 10;
	}
	else
	{
		adjust = T0 * 25;
	}

	ret = rng.RANDOM() + adjust - 127;
	
	if (ret < 0)
	{
		return false;
	}
	else
	{
		return true;
	}
}

// Calculates and assesses damage from a successful attack
bool Player::DAMAGE(int AP, int AMO, int APO,
					int DP, int DMD, int DPD, dodSHORT * DD)
{
	int a;

	a = ((AP * AMO) >> 7);
	a = ((a * DMD) >> 7);
	*DD += (dodSHORT) a;

	a = ((AP * APO) >> 7);
	a = ((a * DPD) >> 7);
	*DD += (dodSHORT) a;

	if ((dodSHORT) DP > *DD)
	{
		return true;
	}
	else
	{
		return false;
	}
}

// Processes CLIMB command
void Player::PCLIMB()
{
	Uint32 ticks1;

	dodBYTE vres;
	int		res;
	dodBYTE	A, B;
	RowCol	rc;
	dodSHORT temp;
	
	rc.setRC(PROW,PCOL);
	vres = dungeon.VFIND(rc);
	if (vres == 255)
	{
		parser.CMDERR();
		return;
	}
	res = parser.PARSER(parser.DIRTAB, &A, &B, true);
	if (res <= 0)
	{
		parser.CMDERR();
		return;
	}
	else
	{
		if (A == 4)
		{
			// Climb Up
			if (vres == 1 || vres == 0)
			{
				viewer.displayPrepare();
				temp = viewer.DSPMOD;
				viewer.DSPMOD = 3;
				viewer.draw_game();
				ticks1 = SDL_GetTicks();
				scheduler.curTime = ticks1;
				do
				{
					if (scheduler.curTime >= scheduler.TCBLND[0].next_time)
					{
						scheduler.CLOCK();
					}
					scheduler.curTime = SDL_GetTicks();
				} while (scheduler.curTime < ticks1 + viewer.prepPause);
				viewer.DSPMOD = temp;
				--game.LEVEL;
				creature.NEWLVL();
				game.INIVU();
			}
			else
			{
				parser.CMDERR();
				return;
			}
		}
		else if (A == 5)
		{
			// Climb Down
			if (vres == 3 || vres == 2)
			{
				viewer.displayPrepare();
				temp = viewer.DSPMOD;
				viewer.DSPMOD = 3;
				viewer.draw_game();
				ticks1 = SDL_GetTicks();
				scheduler.curTime = ticks1;
				do
				{
					if (scheduler.curTime >= scheduler.TCBLND[0].next_time)
					{
						scheduler.CLOCK();
					}
					scheduler.curTime = SDL_GetTicks();
				} while (scheduler.curTime < ticks1 + viewer.prepPause);
				viewer.DSPMOD = temp;
				++game.LEVEL;
				creature.NEWLVL();
				game.INIVU();
			}
			else
			{
				parser.CMDERR();
				return;
			}
		}
		else
		{
			parser.CMDERR();
			return;
		}
	}
}

// Processes DROP command
void Player::PDROP()
{
	int res, idx;
	res = parser.PARHND();
	if (res == -1)
	{
		return;
	}
	if (res == 0 && PLHAND == -1)
	{
		parser.CMDERR();
		return;
	}
	if (res == 1 && PRHAND == -1)
	{
		parser.CMDERR();
		return;
	}

	if (res == 0)
	{
		idx = PLHAND;
		PLHAND = -1;
	}
	else
	{
		idx = PRHAND;
		PRHAND = -1;
	}

	object.OCBLND[idx].P_OCOWN = 0;
	object.OCBLND[idx].P_OCROW = PROW;
	object.OCBLND[idx].P_OCCOL = PCOL;
	object.OCBLND[idx].P_OCLVL = game.LEVEL;

	POBJWT -= object.OBJWGT[object.OCBLND[idx].P_OCCLS];
	HUPDAT();

	viewer.STATUS();
	viewer.PUPDAT();
}

// Processes EXAMINE command
void Player::PEXAM()
{
	viewer.DSPMOD = 2;
	viewer.PUPDAT();
}

// Processes GET command
void Player::PGET()
{
	int res, idx;
	res = parser.PARHND();
	if (res == -1)
	{
		return;
	}
	if (res == 0 && PLHAND != -1)
	{
		parser.CMDERR();
		return;
	}
	if (res == 1 && PRHAND != -1)
	{
		parser.CMDERR();
		return;
	}
	if ( !object.PAROBJ() )
	{
		return;
	}

	bool match = false;
	object.OFINDF = 0;
	do
	{
		idx = object.OFIND(RowCol(PROW, PCOL));
		if (idx == -1)
		{
			parser.CMDERR();
			return;
		}
		if (object.SPEFLG == 0)
		{
			if (object.OBJCLS == object.OCBLND[idx].P_OCCLS)
			{
				match = true;
			}
		}
		else
		{
			if (object.OBJTYP == object.OCBLND[idx].P_OCTYP)
			{
				match = true;
			}
		}
	} while (match == false);

	if (res == 0)
	{
		PLHAND = idx;
	}
	else
	{
		PRHAND = idx;
	}
	++object.OCBLND[idx].P_OCOWN;

	POBJWT += object.OBJWGT[object.OCBLND[idx].P_OCCLS];
	HUPDAT();

	viewer.STATUS();
	viewer.PUPDAT();
}

// Processes INCANT command
void Player::PINCAN()
{
	int		res;
	dodBYTE	A, B;
	RowCol	rc;
	SDL_Event event;
	Uint32 ticks1, ticks2;
	
	res = parser.PARSER(object.ADJTAB, &A, &B, true);
	if (res <= 0)
	{
		return;
	}

	if (parser.FULFLG == 0)
	{
		return;
	}

	object.OBJTYP = A;
	object.OBJCLS = B;

	if (PLHAND != -1)
	{
		if (object.OCBLND[PLHAND].P_OCCLS == 1)
		{
			if (object.OCBLND[PLHAND].P_OCXX1 == object.OBJTYP)
			{
				object.OCBLND[PLHAND].P_OCTYP = object.OBJTYP;
				object.OCBFIL(object.OBJTYP, PLHAND);

				// make ring sound
				Mix_PlayChannel(object.objChannel,
					object.objSound[object.OCBLND[PLHAND].P_OCCLS], 0);
				while (Mix_Playing(object.objChannel) == 1)
				{
					if (scheduler.curTime >= scheduler.TCBLND[0].next_time)
					{
						scheduler.CLOCK();
					}
					scheduler.curTime = SDL_GetTicks();
				}
				
				viewer.STATUS();
				viewer.PUPDAT();
				object.OCBLND[PLHAND].P_OCXX1 = -1;
				if (object.OBJTYP == 0x12)
				{
					// winner
					while(SDL_PollEvent(&event))
					{
						; // clear event buffer
					}

					// Pause so player can see status line
					ticks1 = SDL_GetTicks();
					do
					{
						ticks2 = SDL_GetTicks();
					} while (ticks2 < ticks1 + wizDelay);

					viewer.clearArea(&viewer.TXTSTS);
					viewer.clearArea(&viewer.TXTPRI);
					scheduler.winFadeLoop();
					game.hasWon = true;
				}
				else
				{
					return;
				}
			}
		}
	}

	if (PRHAND != -1)
	{
		if (object.OCBLND[PRHAND].P_OCCLS == 1)
		{
			if (object.OCBLND[PRHAND].P_OCXX1 == object.OBJTYP)
			{
				object.OCBLND[PRHAND].P_OCTYP = object.OBJTYP;
				object.OCBFIL(object.OBJTYP, PRHAND);

				// make ring sound
				Mix_PlayChannel(object.objChannel,
					object.objSound[object.OCBLND[PRHAND].P_OCCLS], 0);
				while (Mix_Playing(object.objChannel) == 1)
				{
					if (scheduler.curTime >= scheduler.TCBLND[0].next_time)
					{
						scheduler.CLOCK();
					}
					scheduler.curTime = SDL_GetTicks();
				}
				
				viewer.STATUS();
				viewer.PUPDAT();
				object.OCBLND[PRHAND].P_OCXX1 = -1;
				if (object.OBJTYP == 0x12)
				{
					// Do winner
					while(SDL_PollEvent(&event))
					{
						; // clear event buffer
					}

					// Pause so player can see status line
					ticks1 = SDL_GetTicks();
					do
					{
						ticks2 = SDL_GetTicks();
					} while (ticks2 < ticks1 + wizDelay);

					viewer.clearArea(&viewer.TXTSTS);
					viewer.clearArea(&viewer.TXTPRI);
					scheduler.winFadeLoop();
					game.hasWon = true;
				}
				else
				{
					return;
				}
			}
		}
	}
}

// Processes LOOK command
void Player::PLOOK()
{
	viewer.DSPMOD = 1;
	viewer.PUPDAT();
}

// Processes MOVE command
void Player::PMOVE()
{
	int		res;
	dodBYTE	A, B;
	Uint32 ticks1;
	
	res = parser.PARSER(parser.DIRTAB, &A, &B, true);
	if (res < 0)
	{
		parser.CMDERR();
		return;
	}
	else if (res == 0)
	{
		// Move Forward
		--viewer.HLFSTP;
		viewer.PUPDAT();
		ticks1 = SDL_GetTicks();
		scheduler.curTime = ticks1;
		do
		{
			if (scheduler.curTime >= scheduler.TCBLND[0].next_time)
			{
				scheduler.CLOCK();
			}
			scheduler.curTime = SDL_GetTicks();
		} while (scheduler.curTime < ticks1 + moveDelay);
		viewer.HLFSTP = 0;
		PSTEP(0);
		PDAM += (POBJWT >> 3) + 3;
		HUPDAT();
		--viewer.UPDATE;
		viewer.draw_game();
		return;
	}
	else
	{
		if (A == 2)
		{
			// Move Back
			--viewer.BAKSTP;
			viewer.PUPDAT();
			ticks1 = SDL_GetTicks();
			scheduler.curTime = ticks1;
			do
			{
				if (scheduler.curTime >= scheduler.TCBLND[0].next_time)
				{
					scheduler.CLOCK();
				}
				scheduler.curTime = SDL_GetTicks();
			} while (scheduler.curTime < ticks1 + moveDelay);
			viewer.BAKSTP = 0;
			PSTEP(2);
			PDAM += (POBJWT / 8) + 3;
			HUPDAT();
			--viewer.UPDATE;
			viewer.draw_game();
			return;
		}
		else if (A == 1)
		{
			// Move Right
			if (PSTEP(1))
			{
				if (viewer.DSPMOD == 1)
				{
					ShowTurn(A);
				}
			}
			PDAM += (POBJWT / 8) + 3;
			HUPDAT();
			--viewer.UPDATE;
			viewer.draw_game();
			return;
		}
		else if (A == 0)
		{
			// Move Left
			if (PSTEP(3))
			{
				if (viewer.DSPMOD == 1)
				{
					ShowTurn(A);
				}
			}
			PDAM += (POBJWT / 8) + 3;
			HUPDAT();
			--viewer.UPDATE;
			viewer.draw_game();
			return;
		}
		else
		{
			parser.CMDERR();
			return;
		}
	}
}

// Processes PULL command
void Player::PPULL()
{
	if (BAGPTR == -1)
	{
		parser.CMDERR();
		return;
	}
	
	int res;
	res = parser.PARHND();
	if (res == -1)
	{
		return;
	}
	if (res == 0 && PLHAND != -1)
	{
		parser.CMDERR();
		return;
	}
	if (res == 1 && PRHAND != -1)
	{
		parser.CMDERR();
		return;
	}
	if ( !object.PAROBJ() )
	{
		return;
	}

	bool onHead = true;
	bool match = false;
	int curPtr, prevPtr;

	do
	{
		if (onHead)
		{
			curPtr = BAGPTR;
		}
		else
		{
			prevPtr = curPtr;
			curPtr = object.OCBLND[curPtr].P_OCPTR;
			if (curPtr == -1)
			{
				parser.CMDERR();
				return;
			}
		}

		if (object.SPEFLG == 0)
		{
			if (object.OCBLND[curPtr].P_OCCLS == object.OBJCLS)
			{
				match = true;
			}
		}
		else
		{
			if (object.OCBLND[curPtr].P_OCTYP == object.OBJTYP)
			{
				match = true;
			}
		}
		if (match)
		{
			break;
		}
		if (onHead)
		{
			onHead = false;
		}
	} while (true);

	if (onHead)
	{
		BAGPTR = object.OCBLND[curPtr].P_OCPTR;
	}
	else
	{
		object.OCBLND[prevPtr].P_OCPTR = object.OCBLND[curPtr].P_OCPTR;
	}

	if (res == 0)
	{
		PLHAND = curPtr;
	}
	else
	{
		PRHAND = curPtr;
	}

	if (curPtr == PTORCH)
	{
		PTORCH = -1;
	}

	viewer.STATUS();
	viewer.PUPDAT();
}

// Processes REVEAL command
void Player::PREVEA()
{
	int res, idx, req;
	res = parser.PARHND();
	if (res == -1)
	{
		return;
	}
	if (res == 0 && PLHAND == -1)
	{
		return;
	}
	if (res == 1 && PRHAND == -1)
	{
		return;
	}
	if (res == 0)
	{
		idx = PLHAND;
	}
	else
	{
		idx = PRHAND;
	}
	req = object.OCBLND[idx].P_OCREV;
	if (req == 0)
	{
		return;
	}
	if (req * 25 > PPOW)
	{
		return;
	}
	object.OCBFIL(object.OCBLND[idx].P_OCTYP, idx);
	object.OCBLND[idx].P_OCREV = 0;
	viewer.STATUS();
}

// Processes STOW command
void Player::PSTOW()
{
	int res;
	res = parser.PARHND();
	if (res == -1)
	{
		return;
	}
	if (res == 0 && PLHAND == -1)
	{
		parser.CMDERR();
		return;
	}
	if (res == 1 && PRHAND == -1)
	{
		parser.CMDERR();
		return;
	}

	if (res == 0)
	{
		object.OCBLND[PLHAND].P_OCPTR = BAGPTR;
		BAGPTR = PLHAND;
		PLHAND = -1;
	}
	else
	{
		object.OCBLND[PRHAND].P_OCPTR = BAGPTR;
		BAGPTR = PRHAND;
		PRHAND = -1;
	}
	viewer.STATUS();
	viewer.PUPDAT();
}

// Processes TURN command
void Player::PTURN()
{
	int		res;
	dodBYTE	A, B;
	
	res = parser.PARSER(parser.DIRTAB, &A, &B, true);
	if (res != 1)
	{
		parser.CMDERR();
		return;
	}
	if (A == 0)
	{
		// Left Turn
		--PDIR;
		PDIR = (PDIR & 3);
		if (viewer.DSPMOD == 1)
		{
			ShowTurn(A);
		}
		--viewer.UPDATE;
		viewer.draw_game();
		return;
	}
	else if (A == 1)
	{
		// Right Turn
		++PDIR;
		PDIR = (PDIR & 3);
		if (viewer.DSPMOD == 1)
		{
			ShowTurn(A);
		}
		--viewer.UPDATE;
		viewer.draw_game();
		return;
	}
	else if (A == 3)
	{
		// About Face
		PDIR += 2;
		PDIR = (PDIR & 3);
		if (viewer.DSPMOD == 1)
		{
			ShowTurn(A);
		}
		--viewer.UPDATE;
		viewer.draw_game();
		return;
	}
	else
	{
		parser.CMDERR();
		return;
	}
}

// Turning Animation
void Player::ShowTurn(dodBYTE A)
{
	int ctr, times, x;
	int offset, dir;
	int inc = 32;
	int lines = 8;
	int y0 = 17;
	int y1 = 135;
	Uint32 ticks1;

	switch (A)
	{
	case 0:
		offset = 8;
		dir = 1;
		times = 1;
		break;
	case 1:
		offset = 248;
		dir = -1;
		times = 1;
		break;
	case 3:
		offset = 248;
		dir = -1;
		times = 2;
		break;
	default:
		break;
	}

	viewer.VXSCAL = 0x80;
	viewer.VYSCAL = 0x80;
	dungeon.RANGE = 0;
	viewer.SETFAD();
	HEARTF = 0;
	glColor3fv(viewer.fgColor);
	for (ctr = 0; ctr < times; ++ctr)
	{
		for (x = 0; x < lines; ++x)
		{
			glClear(GL_COLOR_BUFFER_BIT);
			glLoadIdentity();
			viewer.drawVectorList(viewer.LINES);
			viewer.drawVector((x*inc*dir)+offset,y0,(x*inc*dir)+offset,y1);
			SDL_GL_SwapBuffers();
			ticks1 = SDL_GetTicks();
			scheduler.curTime = ticks1;
			do
			{
				if (scheduler.curTime >= scheduler.TCBLND[0].next_time)
				{
					scheduler.CLOCK();
				}
				scheduler.curTime = SDL_GetTicks();
			} while (scheduler.curTime < ticks1 + turnDelay);
		}
	}
	--HEARTF;
}

// Processes USE command
void Player::PUSE()
{
	int res, idx;
	res = parser.PARHND();
	if (res == -1)
	{
		return;
	}
	if (res == 0 && PLHAND == -1)
	{
		return;
	}
	if (res == 1 && PRHAND == -1)
	{
		return;
	}
	if (res == 0)
	{
		idx = PLHAND;
	}
	else
	{
		idx = PRHAND;
	}

	if (object.OCBLND[idx].P_OCCLS  == object.K_TORC)
	{
		PTORCH = idx;
		if (res == 0)
		{
			object.OCBLND[PLHAND].P_OCPTR = BAGPTR;
			BAGPTR = PLHAND;
			PLHAND = -1;
		}
		else
		{
			object.OCBLND[PRHAND].P_OCPTR = BAGPTR;
			BAGPTR = PRHAND;
			PRHAND = -1;
		}
		viewer.STATUS();
		viewer.PUPDAT();

		// make torch sound
		Mix_PlayChannel(object.objChannel,
			object.objSound[object.OCBLND[idx].P_OCCLS], 0);
		while (Mix_Playing(object.objChannel) == 1)
		{
			if (scheduler.curTime >= scheduler.TCBLND[0].next_time)
			{
				scheduler.CLOCK();
			}
			scheduler.curTime = SDL_GetTicks();
		}
		
		viewer.PUPDAT();
		return;
	}
	else if (object.OCBLND[idx].P_OCTYP  == 5) // Thews Flask
	{
		PPOW += 1000;
		object.OCBLND[idx].P_OCTYP = 0x17;
		object.OCBLND[idx].P_OCREV = 0;
		
		// make flask sound
		Mix_PlayChannel(object.objChannel,
			object.objSound[object.OCBLND[idx].P_OCCLS], 0);
		while (Mix_Playing(object.objChannel) == 1)
		{
			if (scheduler.curTime >= scheduler.TCBLND[0].next_time)
			{
				scheduler.CLOCK();
			}
			scheduler.curTime = SDL_GetTicks();
		}
		
		viewer.STATUS();
		HUPDAT();
	}
	else if (object.OCBLND[idx].P_OCTYP  == 9) // Hale Flask
	{
		PDAM = 0;
		object.OCBLND[idx].P_OCTYP = 0x17;
		object.OCBLND[idx].P_OCREV = 0;
		
		// make flask sound
		Mix_PlayChannel(object.objChannel,
			object.objSound[object.OCBLND[idx].P_OCCLS], 0);
		while (Mix_Playing(object.objChannel) == 1)
		{
			if (scheduler.curTime >= scheduler.TCBLND[0].next_time)
			{
				scheduler.CLOCK();
			}
			scheduler.curTime = SDL_GetTicks();
		}
		
		viewer.STATUS();
		HUPDAT();
	}
	else if (object.OCBLND[idx].P_OCTYP  == 8) // Abye Flask
	{
		PDAM += (short) ((double) PPOW * 0.8);
		object.OCBLND[idx].P_OCTYP = 0x17;
		object.OCBLND[idx].P_OCREV = 0;
		
		// make flask sound
		Mix_PlayChannel(object.objChannel,
			object.objSound[object.OCBLND[idx].P_OCCLS], 0);
		while (Mix_Playing(object.objChannel) == 1)
		{
			if (scheduler.curTime >= scheduler.TCBLND[0].next_time)
			{
				scheduler.CLOCK();
			}
			scheduler.curTime = SDL_GetTicks();
		}
		
		viewer.STATUS();
		HUPDAT();
	}
	else if (object.OCBLND[idx].P_OCTYP  == 4) // Seer Scroll
	{
		viewer.MAPFLG = true;
		if (object.OCBLND[idx].P_OCREV != 0)
		{
			return;
		}
		
		// make scroll sound
		Mix_PlayChannel(object.objChannel,
			object.objSound[object.OCBLND[idx].P_OCCLS], 0);
		while (Mix_Playing(object.objChannel) == 1)
		{
			if (scheduler.curTime >= scheduler.TCBLND[0].next_time)
			{
				scheduler.CLOCK();
			}
			scheduler.curTime = SDL_GetTicks();
		}
		
		HEARTF = 0;
		viewer.DSPMOD = 0;
		viewer.PUPDAT();
		return;
	}
	else if (object.OCBLND[idx].P_OCTYP  == 7) // Vision Scroll
	{
		viewer.MAPFLG = false;
		if (object.OCBLND[idx].P_OCREV != 0)
		{
			return;
		}
		
		// make scroll sound
		Mix_PlayChannel(object.objChannel,
			object.objSound[object.OCBLND[idx].P_OCCLS], 0);
		while (Mix_Playing(object.objChannel) == 1)
		{
			if (scheduler.curTime >= scheduler.TCBLND[0].next_time)
			{
				scheduler.CLOCK();
			}
			scheduler.curTime = SDL_GetTicks();
		}
		
		HEARTF = 0;
		viewer.DSPMOD = 0;
		viewer.PUPDAT();
		return;
	}
}

// Processes ZLOAD command
void Player::PZLOAD()
{
	int tctr = 0;
	int preLen;
	memset(parser.TOKEN, -1, 33);
	memset(oslink.gamefile,0,oslink.gamefileLen);
	strcpy(oslink.gamefile,oslink.savedDir);
	strcat(oslink.gamefile,oslink.pathSep);
	preLen = strlen(oslink.gamefile);
	if (parser.GETTOK())
	{
		strcpy(oslink.gamefile,oslink.savedDir);
		strcat(oslink.gamefile,oslink.pathSep);
		tctr = 0;
		while (parser.TOKEN[tctr] != 0xFF)
		{
			oslink.gamefile[tctr + preLen] = parser.TOKEN[tctr] + 'A' - 1;
			++tctr;
		}
		strcat(oslink.gamefile,".dod");
	}
	else
	{
		strcat(oslink.gamefile,"game.dod");
	}

	if ((oslink.fptr = fopen(oslink.gamefile,"r")) == NULL)
	{
		parser.CMDERR();
		return;
	}
	else
	{
		fclose(oslink.fptr);
		--scheduler.ZFLAG;
		return;
	}
}

// Processes ZSAVE command
void Player::PZSAVE()
{
	int tctr = 0;
	int preLen;
	memset(parser.TOKEN, -1, 33);
	memset(oslink.gamefile,0,oslink.gamefileLen);
	strcpy(oslink.gamefile,oslink.savedDir);
	strcat(oslink.gamefile,oslink.pathSep);
	preLen = strlen(oslink.gamefile);
	if (parser.GETTOK())
	{
		while (parser.TOKEN[tctr] != 0xFF)
		{
			oslink.gamefile[tctr + preLen] = parser.TOKEN[tctr] + 'A' - 1;
			++tctr;
		}
		strcat(oslink.gamefile,".dod");
	}
	else
	{
		strcat(oslink.gamefile,"game.dod");
	}

	++scheduler.ZFLAG;
	return;
}

// Attempts to move player in given direction
bool Player::PSTEP(dodBYTE dir)
{
	dodBYTE B;
	B = dir + PDIR;
	B &= 3;
	if (dungeon.STEPOK(PROW, PCOL, B))
	{
		PROW += dungeon.STPTAB[B * 2];
		PCOL += dungeon.STPTAB[(B * 2) + 1];
		return true;
	}
	else
	{
		// do thud sound
		Mix_PlayChannel(object.objChannel, thud, 0);
		while (Mix_Playing(object.objChannel) == 1)
		{
			if (scheduler.curTime >= scheduler.TCBLND[0].next_time)
			{
				scheduler.CLOCK();
			}
			scheduler.curTime = SDL_GetTicks();
		}
		
		return false;
	}
}

