/****************************************
Daggorath PC-Port Version 0.2.1
Richard Hunerlach
November 13, 2002

The copyright for Dungeons of Daggorath
is held by Douglas J. Morgan.
(c) 1982, DynaMicro
*****************************************/

// Dungeons of Daggorath
// PC-Port
// Filename: sched.cpp
//
// Implementation of the Scheduler class

#include <iostream>
#include <fstream>

using namespace std;

#include "sched.h"
#include "player.h"
#include "viewer.h"
#include "oslink.h"
#include "creature.h"
#include "parser.h"
#include "dodgame.h"
#include "dungeon.h"
#include "rng.h"
#include "object.h"

extern Object	object;
extern RNG		rng;
extern Dungeon	dungeon;
extern dodGame	game;
extern Parser	parser;
extern Creature	creature;
extern Player	player;
extern Viewer	viewer;
extern OS_Link	oslink;

const dodSHORT Scheduler::clockTCB	= 0;
const dodSHORT Scheduler::playerTCB	= 1;
const dodSHORT Scheduler::luknewTCB	= 2;
const dodSHORT Scheduler::hslowTCB	= 3;
const dodSHORT Scheduler::burnerTCB	= 4;
const dodSHORT Scheduler::cregenTCB	= 5;
const dodSHORT Scheduler::cmoveTCB	= 6;

// Constructor
Scheduler::Scheduler() : TCBPTR(0),
						 KBDHDR(0),
						 KBDTAL(0),
						 SLEEP(0),
						 NOISEF(0),
						 NOISEV(0),
						 ZFLAG(0),
						 hrtChannel(0)
{
	int ctr;

	for (ctr = 0; ctr < 38; ++ctr)
	{
		TCBLND[ctr].clear();
	}
}

void Scheduler::Reset()
{
	TCBPTR = 0;
	KBDHDR = 0;
	KBDTAL = 0;
	SLEEP = 0;
	NOISEF = 0;
	NOISEV = 0;
	ZFLAG = 0;
	int ctr;

	for (ctr = 0; ctr < 38; ++ctr)
	{
		TCBLND[ctr].clear();
	}
}

// Public Interface
void Scheduler::LoadSounds()
{
	char fn[20];
	char fpre[7];
	strcpy(fpre,oslink.soundDir);
	strcat(fpre,oslink.pathSep);

	strcpy(fn,fpre);
	strcat(fn,"17_heart.wav");
	hrtSound[0] = Mix_LoadWAV(fn);

	strcpy(fn,fpre);
	strcat(fn,"18_heart.wav");
	hrtSound[1] = Mix_LoadWAV(fn);
}

// Creates initial Task Blocks
void Scheduler::SYSTCB()
{
	int	ctr;
	int	TCBindex;

	for (ctr = 0; ctr < 38; ++ctr)
	{
		TCBLND[ctr].clear();
	}
	
	TCBPTR = 0;

	TCBLND[0].type = clockTCB;
	TCBLND[0].frequency = 17;		// One JIFFY
	TCBindex = GETTCB();

	TCBLND[1].type = playerTCB;
	TCBLND[1].frequency = 17;		// One JIFFY
	TCBindex = GETTCB();

	TCBLND[2].type = luknewTCB;
	TCBLND[2].frequency = 300;		// Three TENTHs
	TCBindex = GETTCB();
	
	TCBLND[3].type = hslowTCB;
	TCBindex = GETTCB();
	
	TCBLND[4].type = burnerTCB;
	TCBLND[4].frequency = 5000;		// Five Seconds
	TCBindex = GETTCB();
	
	TCBLND[5].type = cregenTCB;
	TCBLND[5].frequency = 300000;	// Five minutes
	TCBindex = GETTCB();
}

// This is the Main Loop of the game.  Originally it was
// a port of the scheduling algorithm used in the source
// code, but it has been entirely replaced with a simpler
// algorithm that works better on modern platforms.
//
// It uses milliseconds instead of JIFFYs.  And it does
// not have any queues, but a simple array of Task objects.
void Scheduler::SCHED()
{
	// Initialization
	int	result = 0;  // not currently being used
	int	ctr = 0;

	// Main game execution loop
	do
	{
		curTime = SDL_GetTicks();

		if (curTime >= TCBLND[ctr].next_time)
		{
			switch (TCBLND[ctr].type)
			{
			case clockTCB:
				CLOCK();
				break;
			case playerTCB:
				result = player.PLAYER();
				break;
			case luknewTCB:
				result = viewer.LUKNEW();
				break;
			case hslowTCB:
				result = player.HSLOW();
				break;
			case burnerTCB:
				result = player.BURNER();
				break;
			case cregenTCB:
				result = creature.CREGEN();
				break;
			case cmoveTCB:
				result = creature.CMOVE(ctr, TCBLND[ctr].data);
				break;
			default:
				// error
				break;
			}
		}

		(ctr < TCBPTR) ? ++ctr : ctr = 0;

		if (ZFLAG != 0) // Saving or Loading
		{
			if (ZFLAG == 0xFF)
			{
				return; // Load game abandons current game
			}
			else
			{
				SAVE();
				ZFLAG = 0;
			}
		}

		if (player.PLRBLK.P_ATPOW < player.PLRBLK.P_ATDAM)
		{
			return; // Death
		}

		if (game.hasWon)
		{
			return; // Victory
		}

	} while (result == 0);
}

// This is the heart of the game, literally.  It manages
// the heartbeat, calls for the screen to be redrawn, and
// polls the OS for key strokes.
//
// This routine is called every 17 milliseconds from the
// scheduler, and also from the blocking loops after each
// sound, which allows the heartbeat to mix in with the
// other sounds.
void Scheduler::CLOCK()
{
	// Update Task's next_time
	TCBLND[0].next_time = curTime +
		TCBLND[0].frequency;

	// Update elaplsed time
	elapsedTime = curTime - TCBLND[0].prev_time;
	
	// Reality check
	if (elapsedTime > 126 * 17)
	{
		elapsedTime = 126 * 17;
	}

	// Update Task's prev_time
	TCBLND[0].prev_time = curTime;
	
	if (player.HBEATF != 0)
	{
		player.HEARTC -= (elapsedTime / 17);
		if ((player.HEARTC & 0x80) != 0)
		{
			player.HEARTC = 0;
		}
		if (player.HEARTC == 0)
		{
			player.HEARTC = player.HEARTR;
			
			// make sound
			Mix_PlayChannel(hrtChannel, hrtSound[(dodBYTE) (player.HEARTS + 1)], 0);
			while (Mix_Playing(hrtChannel) == 1) ; // !!!

			if (player.HEARTF != 0)
			{
				if ((player.HEARTS & 0x80) != 0)
				{
					// small
					viewer.statArea[15] = '<';
					viewer.statArea[16] = '>';
					player.HEARTS = 0;
				}
				else
				{
					// large
					viewer.statArea[15] = '{';
					viewer.statArea[16] = '}';
					player.HEARTS = -1;
				}
				--viewer.UPDATE;
				viewer.draw_game();
			}
		}
	}

	if (player.FAINT == 0)
	{
		// Perform Keyboard Input
		oslink.process_events();
	}
}

// Gets next available Task Block and updates the index
int Scheduler::GETTCB()
{
	++TCBPTR;
	return (TCBPTR - 1);
}

// All the following methods should really be moved to the
// OS_Link class.

// Used by wizard fade in/out function
bool Scheduler::fadeLoop()
{
	SDL_Event event;
	viewer.displayCopyright();
	viewer.displayWelcomeMessage();

	// Start buzz
	Mix_Volume(viewer.fadChannel, 0);
	Mix_PlayChannel(viewer.fadChannel, creature.buzz, -1);
	
	while(true)
	{
		if ( keyCheck() )
		{
			viewer.clearArea(&viewer.TXTPRI);
			while(SDL_PollEvent(&event))
			{
				; // clear event buffer
			}

			// Stop buzz
			Mix_HaltChannel(viewer.fadChannel);

			return false;	// auto-play mode off == start demo game
		}
		if ( viewer.draw_fade() )
		{
			// Stop buzz
			Mix_HaltChannel(viewer.fadChannel);

			return true;	// auto-play mode on == start regular game
		}
	}
}

void Scheduler::deathFadeLoop()
{
	SDL_Event event;
	viewer.displayDeath();
	viewer.fadeVal = -2;
	viewer.VXSCAL = 0x80;
	viewer.VYSCAL = 0x80;
	viewer.delay = 0;
	viewer.VCTFAD = 32;
	viewer.done = false;

	while(SDL_PollEvent(&event))
	{
		; // clear event buffer
	}

	// Start buzz
	Mix_Volume(viewer.fadChannel, 0);
	Mix_PlayChannel(viewer.fadChannel, creature.buzz, -1);
	
	while (!viewer.done)
	{
		viewer.death_fade(viewer.W1_VLA);
		EscCheck();
	}

	// Stop buzz
	Mix_HaltChannel(viewer.fadChannel);

	while(SDL_PollEvent(&event))
	{
		; // clear event buffer
	}

	while(true)
	{
		viewer.death_fade(viewer.W1_VLA);
		if ( keyCheck() )
		{
			viewer.clearArea(&viewer.TXTPRI);
			while(SDL_PollEvent(&event))
			{
				; // clear event buffer
			}
			return;
		}
	}
}

void Scheduler::winFadeLoop()
{
	SDL_Event event;
	bool loopDone = false;
	viewer.displayWinner();
	viewer.fadeVal = -2;
	viewer.VXSCAL = 0x80;
	viewer.VYSCAL = 0x80;
	viewer.delay = 0;
	viewer.VCTFAD = 32;
	viewer.done = false;

	while(SDL_PollEvent(&event))
	{
		; // clear event buffer
	}

	// Start buzz
	Mix_Volume(viewer.fadChannel, 0);
	Mix_PlayChannel(viewer.fadChannel, creature.buzz, -1);
	
	while (!viewer.done)
	{
		viewer.death_fade(viewer.W2_VLA);
		EscCheck();
	}

	// Stop buzz
	Mix_HaltChannel(viewer.fadChannel);

	while(true)
	{
		viewer.death_fade(viewer.W2_VLA);
		if ( keyCheck() )
		{
			viewer.clearArea(&viewer.TXTPRI);
			while(SDL_PollEvent(&event))
			{
				; // clear event buffer
			}
			return;
		}
	}

	while(SDL_PollEvent(&event))
	{
		; // clear event buffer
	}
}

// Used by wizard fade in/out function
bool Scheduler::keyCheck()
{
	SDL_Event event;
	while(SDL_PollEvent(&event))
	{
		switch(event.type)
		{
		case SDL_KEYDOWN:
			return ( keyHandler(&event.key.keysym) );
			break;
		case SDL_QUIT:
			oslink.quitSDL(0); // eventually change to meta-menu
			break;
		}
	}
	return false;
}

// Used by wizard fade in/out function
bool Scheduler::keyHandler(SDL_keysym * keysym)
{
	switch(keysym->sym)
	{
	case SDLK_ESCAPE:
		oslink.quitSDL(0); // eventually change to meta-menu
		return false;
	default:
		return true;
	}
}

// Used during fainting and intermission
bool Scheduler::EscCheck()
{
	SDL_Event event;
	while(SDL_PollEvent(&event))
	{
		switch(event.type)
		{
		case SDL_KEYDOWN:
			return ( keyHandler(&event.key.keysym) );
			break;
		case SDL_QUIT:
			oslink.quitSDL(0); // eventually change to meta-menu
			break;
		}
	}
	return false;
}

// Used by wizard fade in/out function
bool Scheduler::EscHandler(SDL_keysym * keysym)
{
	switch(keysym->sym)
	{
	case SDLK_ESCAPE:
		oslink.quitSDL(0); // eventually change to meta-menu
		return true;
	default:
		return false;
	}
}

void Scheduler::SAVE()
{
	ofstream	fout;
	int			ctr;
	char 		outstr[64];
	dodBYTE		de[15] = { 4, 9, 19, 11, 0, 5, 18, 18, 15,
						   18, 0, 27, 27, 27, 255 };

	fout.open(oslink.gamefile, ios::trunc);
	if (!fout)
	{
		// DISK ERROR
		viewer.OUTSTR(de);
		viewer.PROMPT();
		return;
	}

	sprintf(outstr, "%d", game.LEVEL); fout << outstr << endl;
	sprintf(outstr, "%d", dungeon.VFTPTR); fout << outstr << endl;

	for (ctr = 0; ctr < 1024; ++ctr)
	{
		sprintf(outstr, "%d", dungeon.MAZLND[ctr]); fout << outstr << endl;
	}

	sprintf(outstr, "%d", player.PROW); fout << outstr << endl;
	sprintf(outstr, "%d", player.PCOL); fout << outstr << endl;
	sprintf(outstr, "%d", player.POBJWT); fout << outstr << endl;
	sprintf(outstr, "%d", player.PPOW); fout << outstr << endl;
	sprintf(outstr, "%d", player.PLHAND); fout << outstr << endl;
	sprintf(outstr, "%d", player.PRHAND); fout << outstr << endl;
	sprintf(outstr, "%d", player.PDAM); fout << outstr << endl;
	sprintf(outstr, "%d", player.PDIR); fout << outstr << endl;
	sprintf(outstr, "%d", player.PTORCH); fout << outstr << endl;
	sprintf(outstr, "%d", player.PRLITE); fout << outstr << endl;
	sprintf(outstr, "%d", player.PMLITE); fout << outstr << endl;
	sprintf(outstr, "%d", player.FAINT); fout << outstr << endl;
	sprintf(outstr, "%d", player.BAGPTR); fout << outstr << endl;
	sprintf(outstr, "%d", player.HEARTF); fout << outstr << endl;
	sprintf(outstr, "%d", player.HEARTC); fout << outstr << endl;
	sprintf(outstr, "%d", player.HEARTR); fout << outstr << endl;
	sprintf(outstr, "%d", player.HEARTS); fout << outstr << endl;
	sprintf(outstr, "%d", player.HBEATF); fout << outstr << endl;
	sprintf(outstr, "%d", rng.SEED[0]); fout << outstr << endl;
	sprintf(outstr, "%d", rng.SEED[1]); fout << outstr << endl;
	sprintf(outstr, "%d", rng.SEED[2]); fout << outstr << endl;
	sprintf(outstr, "%d", rng.carry); fout << outstr << endl;
	sprintf(outstr, "%d", creature.FRZFLG); fout << outstr << endl;
	sprintf(outstr, "%d", creature.CMXPTR); fout << outstr << endl;

	for (ctr = 0; ctr < 60; ++ctr)
	{
		sprintf(outstr, "%d", creature.CMXLND[ctr]); fout << outstr << endl;
	}

	for (ctr = 0; ctr < 32; ++ctr)
	{
		sprintf(outstr, "%d", creature.CCBLND[ctr].P_CCPOW); fout << outstr << endl;
		sprintf(outstr, "%d", creature.CCBLND[ctr].P_CCMGO); fout << outstr << endl;
		sprintf(outstr, "%d", creature.CCBLND[ctr].P_CCMGD); fout << outstr << endl;
		sprintf(outstr, "%d", creature.CCBLND[ctr].P_CCPHO); fout << outstr << endl;
		sprintf(outstr, "%d", creature.CCBLND[ctr].P_CCPHD); fout << outstr << endl;
		sprintf(outstr, "%d", creature.CCBLND[ctr].P_CCTMV); fout << outstr << endl;
		sprintf(outstr, "%d", creature.CCBLND[ctr].P_CCTAT); fout << outstr << endl;
		sprintf(outstr, "%d", creature.CCBLND[ctr].P_CCOBJ); fout << outstr << endl;
		sprintf(outstr, "%d", creature.CCBLND[ctr].P_CCDAM); fout << outstr << endl;
		sprintf(outstr, "%d", creature.CCBLND[ctr].P_CCUSE); fout << outstr << endl;
		sprintf(outstr, "%d", creature.CCBLND[ctr].P_CCTYP); fout << outstr << endl;
		sprintf(outstr, "%d", creature.CCBLND[ctr].P_CCDIR); fout << outstr << endl;
		sprintf(outstr, "%d", creature.CCBLND[ctr].P_CCROW); fout << outstr << endl;
		sprintf(outstr, "%d", creature.CCBLND[ctr].P_CCCOL); fout << outstr << endl;
	}

	sprintf(outstr, "%d", object.OFINDF); fout << outstr << endl;
	sprintf(outstr, "%d", object.OCBPTR); fout << outstr << endl;
	sprintf(outstr, "%d", object.OFINDP); fout << outstr << endl;

	for (ctr = 0; ctr < 72; ++ctr)
	{
		sprintf(outstr, "%d", object.OCBLND[ctr].P_OCPTR); fout << outstr << endl;
		sprintf(outstr, "%d", object.OCBLND[ctr].P_OCROW); fout << outstr << endl;
		sprintf(outstr, "%d", object.OCBLND[ctr].P_OCCOL); fout << outstr << endl;
		sprintf(outstr, "%d", object.OCBLND[ctr].P_OCLVL); fout << outstr << endl;
		sprintf(outstr, "%d", object.OCBLND[ctr].P_OCOWN); fout << outstr << endl;
		sprintf(outstr, "%d", object.OCBLND[ctr].P_OCXX0); fout << outstr << endl;
		sprintf(outstr, "%d", object.OCBLND[ctr].P_OCXX1); fout << outstr << endl;
		sprintf(outstr, "%d", object.OCBLND[ctr].P_OCXX2); fout << outstr << endl;
		sprintf(outstr, "%d", object.OCBLND[ctr].P_OCTYP); fout << outstr << endl;
		sprintf(outstr, "%d", object.OCBLND[ctr].P_OCCLS); fout << outstr << endl;
		sprintf(outstr, "%d", object.OCBLND[ctr].P_OCREV); fout << outstr << endl;
		sprintf(outstr, "%d", object.OCBLND[ctr].P_OCMGO); fout << outstr << endl;
		sprintf(outstr, "%d", object.OCBLND[ctr].P_OCPHO); fout << outstr << endl;
	}

	sprintf(outstr, "%d", parser.LINPTR); fout << outstr << endl;
	sprintf(outstr, "%d", parser.PARFLG); fout << outstr << endl;
	sprintf(outstr, "%d", parser.PARCNT); fout << outstr << endl;
	sprintf(outstr, "%d", parser.VERIFY); fout << outstr << endl;
	sprintf(outstr, "%d", parser.FULFLG); fout << outstr << endl;
	sprintf(outstr, "%d", parser.KBDHDR); fout << outstr << endl;
	sprintf(outstr, "%d", parser.KBDTAL); fout << outstr << endl;
	sprintf(outstr, "%d", parser.BUFFLG); fout << outstr << endl;
	sprintf(outstr, "%d", parser.LINEND); fout << outstr << endl;
	sprintf(outstr, "%d", parser.TOKEND); fout << outstr << endl;

	for (ctr = 0; ctr < 33; ++ctr)
	{
		sprintf(outstr, "%d", parser.KBDBUF[ctr]); fout << outstr << endl;
		sprintf(outstr, "%d", parser.LINBUF[ctr]); fout << outstr << endl;
		sprintf(outstr, "%d", parser.TOKEN[ctr]); fout << outstr << endl;
		sprintf(outstr, "%d", parser.OBJSTR[ctr]); fout << outstr << endl;
	}

	for (ctr = 0; ctr < 35; ++ctr)
	{
		sprintf(outstr, "%d", parser.STRING[ctr]); fout << outstr << endl;
	}

	for (ctr = 0; ctr < 38; ++ctr)
	{
		sprintf(outstr, "%d", TCBLND[ctr].type); fout << outstr << endl;
		sprintf(outstr, "%d", TCBLND[ctr].data); fout << outstr << endl;
		sprintf(outstr, "%d", TCBLND[ctr].frequency); fout << outstr << endl;
		sprintf(outstr, "%d", TCBLND[ctr].count); fout << outstr << endl;
	}

	sprintf(outstr, "%d", TCBPTR); fout << outstr << endl;
	sprintf(outstr, "%d", KBDHDR); fout << outstr << endl;
	sprintf(outstr, "%d", KBDTAL); fout << outstr << endl;
	sprintf(outstr, "%d", SLEEP); fout << outstr << endl;
	sprintf(outstr, "%d", NOISEF); fout << outstr << endl;
	sprintf(outstr, "%d", NOISEV); fout << outstr << endl;

	sprintf(outstr, "%d", viewer.NEWLUK); fout << outstr << endl;
	sprintf(outstr, "%d", viewer.VCTFAD); fout << outstr << endl;
	sprintf(outstr, "%d", viewer.FADFLG); fout << outstr << endl;
	sprintf(outstr, "%d", viewer.UPDATE); fout << outstr << endl;
	sprintf(outstr, "%d", viewer.DSPMOD); fout << outstr << endl;

	for (ctr = 0; ctr < ((32*4)+1); ++ctr)
	{
		sprintf(outstr, "%d", viewer.textArea[ctr]); fout << outstr << endl;
	}

	for (ctr = 0; ctr < ((32*19)+1); ++ctr)
	{
		sprintf(outstr, "%d", viewer.examArea[ctr]); fout << outstr << endl;
	}

	for (ctr = 0; ctr < ((32*1)+1); ++ctr)
	{
		sprintf(outstr, "%d", viewer.statArea[ctr]); fout << outstr << endl;
	}

	sprintf(outstr, "%d", viewer.TXTPRI.caret); fout << outstr << endl;
	sprintf(outstr, "%d", viewer.TXTPRI.len); fout << outstr << endl;
	sprintf(outstr, "%d", viewer.TXTPRI.top); fout << outstr << endl;

	sprintf(outstr, "%d", viewer.TXTEXA.caret); fout << outstr << endl;
	sprintf(outstr, "%d", viewer.TXTEXA.len); fout << outstr << endl;
	sprintf(outstr, "%d", viewer.TXTEXA.top); fout << outstr << endl;

	sprintf(outstr, "%d", viewer.TXTSTS.caret); fout << outstr << endl;
	sprintf(outstr, "%d", viewer.TXTSTS.len); fout << outstr << endl;
	sprintf(outstr, "%d", viewer.TXTSTS.top); fout << outstr << endl;

	sprintf(outstr, "%d", viewer.RLIGHT); fout << outstr << endl;
	sprintf(outstr, "%d", viewer.MLIGHT); fout << outstr << endl;
	sprintf(outstr, "%d", viewer.OLIGHT); fout << outstr << endl;
	sprintf(outstr, "%d", viewer.VXSCAL); fout << outstr << endl;
	sprintf(outstr, "%d", viewer.VYSCAL); fout << outstr << endl;
	sprintf(outstr, "%d", viewer.TXBFLG); fout << outstr << endl;
	sprintf(outstr, "%d", viewer.tcaret); fout << outstr << endl;
	sprintf(outstr, "%d", viewer.tlen); fout << outstr << endl;
	sprintf(outstr, "%d", viewer.NEWLIN); fout << outstr << endl;

	fout.close();
}

void Scheduler::LOAD()
{
	ifstream	fin;
	int			ctr, in;
	char 		instr[64];
	dodBYTE		de[15] = { 4, 9, 19, 11, 0, 5, 18, 18, 15,
						   18, 0, 27, 27, 27, 255 };

	fin.open(oslink.gamefile);
	if (!fin)
	{
		// DISK ERROR
		viewer.OUTSTR(de);
		viewer.PROMPT();
		return;
	}

	fin >> instr; sscanf(instr, "%d", &in); game.LEVEL = in;
	fin >> instr; sscanf(instr, "%d", &in); dungeon.VFTPTR = in;

	for (ctr = 0; ctr < 1024; ++ctr)
	{
		fin >> instr; sscanf(instr, "%d", &in); dungeon.MAZLND[ctr] = in;
	}

	fin >> instr; sscanf(instr,"%d", &in); player.PROW = in;
	fin >> instr; sscanf(instr,"%d", &in); player.PCOL = in;
	fin >> instr; sscanf(instr,"%d", &in); player.POBJWT = in;
	fin >> instr; sscanf(instr,"%d", &in); player.PPOW = in;
	fin >> instr; sscanf(instr,"%d", &in); player.PLHAND = in;
	fin >> instr; sscanf(instr,"%d", &in); player.PRHAND = in;
	fin >> instr; sscanf(instr,"%d", &in); player.PDAM = in;
	fin >> instr; sscanf(instr,"%d", &in); player.PDIR = in;
	fin >> instr; sscanf(instr,"%d", &in); player.PTORCH = in;
	fin >> instr; sscanf(instr,"%d", &in); player.PRLITE = in;
	fin >> instr; sscanf(instr,"%d", &in); player.PMLITE = in;
	fin >> instr; sscanf(instr,"%d", &in); player.FAINT = in;
	fin >> instr; sscanf(instr,"%d", &in); player.BAGPTR = in;
	fin >> instr; sscanf(instr,"%d", &in); player.HEARTF = in;
	fin >> instr; sscanf(instr,"%d", &in); player.HEARTC = in;
	fin >> instr; sscanf(instr,"%d", &in); player.HEARTR = in;
	fin >> instr; sscanf(instr,"%d", &in); player.HEARTS = in;
	fin >> instr; sscanf(instr,"%d", &in); player.HBEATF = in;
	fin >> instr; sscanf(instr,"%d", &in); rng.SEED[0] = in;
	fin >> instr; sscanf(instr,"%d", &in); rng.SEED[1] = in;
	fin >> instr; sscanf(instr,"%d", &in); rng.SEED[2] = in;
	fin >> instr; sscanf(instr,"%d", &in); rng.carry = in;
	fin >> instr; sscanf(instr,"%d", &in); creature.FRZFLG = in;
	fin >> instr; sscanf(instr,"%d", &in); creature.CMXPTR = in;

	for (ctr = 0; ctr < 60; ++ctr)
	{
		fin >> instr; sscanf(instr,"%d", &in); creature.CMXLND[ctr] = in;
	}

	for (ctr = 0; ctr < 32; ++ctr)
	{
		fin >> instr; sscanf(instr,"%d", &in); creature.CCBLND[ctr].P_CCPOW = in;
		fin >> instr; sscanf(instr,"%d", &in); creature.CCBLND[ctr].P_CCMGO = in;
		fin >> instr; sscanf(instr,"%d", &in); creature.CCBLND[ctr].P_CCMGD = in;
		fin >> instr; sscanf(instr,"%d", &in); creature.CCBLND[ctr].P_CCPHO = in;
		fin >> instr; sscanf(instr,"%d", &in); creature.CCBLND[ctr].P_CCPHD = in;
		fin >> instr; sscanf(instr,"%d", &in); creature.CCBLND[ctr].P_CCTMV = in;
		fin >> instr; sscanf(instr,"%d", &in); creature.CCBLND[ctr].P_CCTAT = in;
		fin >> instr; sscanf(instr,"%d", &in); creature.CCBLND[ctr].P_CCOBJ = in;
		fin >> instr; sscanf(instr,"%d", &in); creature.CCBLND[ctr].P_CCDAM = in;
		fin >> instr; sscanf(instr,"%d", &in); creature.CCBLND[ctr].P_CCUSE = in;
		fin >> instr; sscanf(instr,"%d", &in); creature.CCBLND[ctr].P_CCTYP = in;
		fin >> instr; sscanf(instr,"%d", &in); creature.CCBLND[ctr].P_CCDIR = in;
		fin >> instr; sscanf(instr,"%d", &in); creature.CCBLND[ctr].P_CCROW = in;
		fin >> instr; sscanf(instr,"%d", &in); creature.CCBLND[ctr].P_CCCOL = in;
	}

	fin >> instr; sscanf(instr,"%d", &in); object.OFINDF = in;
	fin >> instr; sscanf(instr,"%d", &in); object.OCBPTR = in;
	fin >> instr; sscanf(instr,"%d", &in); object.OFINDP = in;

	for (ctr = 0; ctr < 72; ++ctr)
	{
		fin >> instr; sscanf(instr,"%d", &in); object.OCBLND[ctr].P_OCPTR = in;
		fin >> instr; sscanf(instr,"%d", &in); object.OCBLND[ctr].P_OCROW = in;
		fin >> instr; sscanf(instr,"%d", &in); object.OCBLND[ctr].P_OCCOL = in;
		fin >> instr; sscanf(instr,"%d", &in); object.OCBLND[ctr].P_OCLVL = in;
		fin >> instr; sscanf(instr,"%d", &in); object.OCBLND[ctr].P_OCOWN = in;
		fin >> instr; sscanf(instr,"%d", &in); object.OCBLND[ctr].P_OCXX0 = in;
		fin >> instr; sscanf(instr,"%d", &in); object.OCBLND[ctr].P_OCXX1 = in;
		fin >> instr; sscanf(instr,"%d", &in); object.OCBLND[ctr].P_OCXX2 = in;
		fin >> instr; sscanf(instr,"%d", &in); object.OCBLND[ctr].P_OCTYP = in;
		fin >> instr; sscanf(instr,"%d", &in); object.OCBLND[ctr].P_OCCLS = in;
		fin >> instr; sscanf(instr,"%d", &in); object.OCBLND[ctr].P_OCREV = in;
		fin >> instr; sscanf(instr,"%d", &in); object.OCBLND[ctr].P_OCMGO = in;
		fin >> instr; sscanf(instr,"%d", &in); object.OCBLND[ctr].P_OCPHO = in;
	}

	fin >> instr; sscanf(instr,"%d", &in); parser.LINPTR = in;
	fin >> instr; sscanf(instr,"%d", &in); parser.PARFLG = in;
	fin >> instr; sscanf(instr,"%d", &in); parser.PARCNT = in;
	fin >> instr; sscanf(instr,"%d", &in); parser.VERIFY = in;
	fin >> instr; sscanf(instr,"%d", &in); parser.FULFLG = in;
	fin >> instr; sscanf(instr,"%d", &in); parser.KBDHDR = in;
	fin >> instr; sscanf(instr,"%d", &in); parser.KBDTAL = in;
	fin >> instr; sscanf(instr,"%d", &in); parser.BUFFLG = in;
	fin >> instr; sscanf(instr,"%d", &in); parser.LINEND = in;
	fin >> instr; sscanf(instr,"%d", &in); parser.TOKEND = in;

	for (ctr = 0; ctr < 33; ++ctr)
	{
		fin >> instr; sscanf(instr,"%d", &in); parser.KBDBUF[ctr] = in;
		fin >> instr; sscanf(instr,"%d", &in); parser.LINBUF[ctr] = in;
		fin >> instr; sscanf(instr,"%d", &in); parser.TOKEN[ctr] = in;
		fin >> instr; sscanf(instr,"%d", &in); parser.OBJSTR[ctr] = in;
	}

	for (ctr = 0; ctr < 35; ++ctr)
	{
		fin >> instr; sscanf(instr,"%d", &in); parser.STRING[ctr] = in;
	}

	for (ctr = 0; ctr < 38; ++ctr)
	{
		fin >> instr; sscanf(instr,"%d", &in); TCBLND[ctr].type = in;
		fin >> instr; sscanf(instr,"%d", &in); TCBLND[ctr].data = in;
		fin >> instr; sscanf(instr,"%d", &in); TCBLND[ctr].frequency = in;
		fin >> instr; sscanf(instr,"%d", &in); TCBLND[ctr].count = in;
	}

	fin >> instr; sscanf(instr,"%d", &in); TCBPTR = in;
	fin >> instr; sscanf(instr,"%d", &in); KBDHDR = in;
	fin >> instr; sscanf(instr,"%d", &in); KBDTAL = in;
	fin >> instr; sscanf(instr,"%d", &in); SLEEP = in;
	fin >> instr; sscanf(instr,"%d", &in); NOISEF = in;
	fin >> instr; sscanf(instr,"%d", &in); NOISEV = in;

	fin >> instr; sscanf(instr,"%d", &in); viewer.NEWLUK = in;
	fin >> instr; sscanf(instr,"%d", &in); viewer.VCTFAD = in;
	fin >> instr; sscanf(instr,"%d", &in); viewer.FADFLG = in;
	fin >> instr; sscanf(instr,"%d", &in); viewer.UPDATE = in;
	fin >> instr; sscanf(instr,"%d", &in); viewer.DSPMOD = in;

	for (ctr = 0; ctr < ((32*4)+1); ++ctr)
	{
		fin >> instr; sscanf(instr,"%d", &in); viewer.textArea[ctr] = in;
	}

	for (ctr = 0; ctr < ((32*19)+1); ++ctr)
	{
		fin >> instr; sscanf(instr,"%d", &in); viewer.examArea[ctr] = in;
	}

	for (ctr = 0; ctr < ((32*1)+1); ++ctr)
	{
		fin >> instr; sscanf(instr,"%d", &in); viewer.statArea[ctr] = in;
	}

	fin >> instr; sscanf(instr,"%d", &in); viewer.TXTPRI.caret = in;
	fin >> instr; sscanf(instr,"%d", &in); viewer.TXTPRI.len = in;
	fin >> instr; sscanf(instr,"%d", &in); viewer.TXTPRI.top = in;

	fin >> instr; sscanf(instr,"%d", &in); viewer.TXTEXA.caret = in;
	fin >> instr; sscanf(instr,"%d", &in); viewer.TXTEXA.len = in;
	fin >> instr; sscanf(instr,"%d", &in); viewer.TXTEXA.top = in;

	fin >> instr; sscanf(instr,"%d", &in); viewer.TXTSTS.caret = in;
	fin >> instr; sscanf(instr,"%d", &in); viewer.TXTSTS.len = in;
	fin >> instr; sscanf(instr,"%d", &in); viewer.TXTSTS.top = in;

	fin >> instr; sscanf(instr,"%d", &in); viewer.RLIGHT = in;
	fin >> instr; sscanf(instr,"%d", &in); viewer.MLIGHT = in;
	fin >> instr; sscanf(instr,"%d", &in); viewer.OLIGHT = in;
	fin >> instr; sscanf(instr,"%d", &in); viewer.VXSCAL = in;
	fin >> instr; sscanf(instr,"%d", &in); viewer.VYSCAL = in;
	fin >> instr; sscanf(instr,"%d", &in); viewer.TXBFLG = in;
	fin >> instr; sscanf(instr,"%d", &in); viewer.tcaret = in;
	fin >> instr; sscanf(instr,"%d", &in); viewer.tlen = in;
	fin >> instr; sscanf(instr,"%d", &in); viewer.NEWLIN = in;

	fin.close();
}

