/****************************************
Daggorath PC-Port Version 0.2.1
Richard Hunerlach
November 13, 2002

The copyright for Dungeons of Daggorath
is held by Douglas J. Morgan.
(c) 1982, DynaMicro
*****************************************/

// Dungeons of Daggorath
// PC-Port
// Filename: viewer.cpp
//
// Implementation of the Viewer class

#include "viewer.h"
#include "oslink.h"
#include "player.h"
#include "sched.h"
#include "parser.h"
#include "object.h"
#include "dungeon.h"
#include "creature.h"

extern Creature		creature;
extern Dungeon		dungeon;
extern Object		object;
extern Parser		parser;
extern Coordinate	crd;
extern OS_Link		oslink;
extern Player		player;
extern Scheduler	scheduler;

// Constructor
Viewer::Viewer() : VCNTRX(128), VCNTRY(76), MAPFLG(true),
				   VDGINV(false), UPDATE(0), DSPMOD(3),
				   HLFSTP(0), BAKSTP(0), MAGFLG(0), PASFLG(0),
				   NORSCL(0), HLFSCL(10), BAKSCL(11), TXBFLG(0),
				   TXB_U(0), tcaret(0), tlen(0), RLIGHT(0), MLIGHT(0),
				   fadChannel(3), buzzStep(300), midPause(2500),
				   prepPause(2500)
{
	VCTFAD = 32;
	initVectorLists();
	delay = 0;
	done = false;
	fadeVal = -2;

	bgColor[0] = 0.0;
	bgColor[1] = 0.0;
	bgColor[2] = 0.0;

	fgColor[0] = 1.0;
	fgColor[1] = 1.0;
	fgColor[2] = 1.0;

	FADFLG = 0;

	TXTPRI.area = textArea;
	TXTPRI.caret = 0;
	TXTPRI.len = 32 * 4;
	TXTPRI.top = 20;

	TXTEXA.area = examArea;
	TXTEXA.caret = 0;
	TXTEXA.len = 32 * 19;
	TXTEXA.top = 0;

	TXTSTS.area = statArea;
	TXTSTS.caret = 0;
	TXTSTS.len = 32 * 1;
	TXTSTS.top = 19;

	LArch[0] = LPAS_VLA;
	LArch[1] = LDOR_VLA;
	LArch[2] = LSD_VLA;
	LArch[3] = LWAL_VLA;

	FArch[0] = FPAS_VLA;
	FArch[1] = FDOR_VLA;
	FArch[2] = FSD_VLA;
	FArch[3] = FWAL_VLA;

	RArch[0] = RPAS_VLA;
	RArch[1] = RDOR_VLA;
	RArch[2] = RSD_VLA;
	RArch[3] = RWAL_VLA;

	FLATAB[0] = 3;
	FLATAB[1] = 0;
	FLATAB[2] = 1;

	FLATABv[0] = LArch;
	FLATABv[1] = FArch;
	FLATABv[2] = RArch;

	FWDOBJ[0] = FLAS_VLA;
	FWDOBJ[1] = RING_VLA;
	FWDOBJ[2] = SCRO_VLA;
	FWDOBJ[3] = SHIE_VLA;
	FWDOBJ[4] = SWOR_VLA;
	FWDOBJ[5] = TORC_VLA;

	FWDCRE[0] = SP_VLA;
	FWDCRE[1] = VI_VLA;
	FWDCRE[2] = S1_VLA;
	FWDCRE[3] = BL_VLA;
	FWDCRE[4] = K1_VLA;
	FWDCRE[5] = S2_VLA;
	FWDCRE[6] = SC_VLA;
	FWDCRE[7] = K2_VLA;
	FWDCRE[8] = WR_VLA;
	FWDCRE[9] = GL_VLA;
	FWDCRE[10] = W0_VLA;
	FWDCRE[11] = W1_VLA;

	Utils::LoadFromHex("C88050321F140C080402FF9C6441281A100A060301", Scale);
	Utils::LoadFromHex("FFC0573EA746C0905132281E6051099820C0E7DEF0", enough1);
	Utils::LoadFromHex("E8000848B00C8A0A3C0D29680A232023DEDDEF60", enough2);
	Utils::LoadFromHex("FFC4543D84D80859D12EC80370A693051050202E20", winner1);
	Utils::LoadFromHex("C80000000003CC0081C5B82E9D0644F7BC", winner2);
	Utils::LoadFromHex("FFC192D00173E882C8047966073E809169593BDEF0", death);
	Utils::LoadFromHex("F8DF0CC92745000265C10352393C0068DACC630948", copyright);
	Utils::LoadFromHex("9FD2020645064A02BA8597BDEF80", welcome1);
	Utils::LoadFromHex("F7BDEA20A0255C72BDD303CC0204E77C83446F7B", welcome2);
	Utils::LoadFromHex("3C24580645D8", prepstr);
	Utils::LoadFromHex("625C0A2133049EF6FC", exam1);
	Utils::LoadFromHex("56C72286959177F0", exam2);
	Utils::LoadFromHex("408235C0235FC0", exam3);
	Utils::LoadFromHex("16F7B0", exps);

	LINES[0] = 2;
	LINES[1] = 2;
	LINES[2] = 0;
	LINES[3] = 16;
	LINES[4] = 255;
	LINES[5] = 16;
	LINES[6] = 2;
	LINES[7] = 0;
	LINES[8] = 136;
	LINES[9] = 255;
	LINES[10] = 136;
}

void Viewer::Reset()
{
	MAPFLG = true;
	VDGINV = false;
	setVid();
	UPDATE = 0;
	DSPMOD = 3;
	HLFSTP = 0;
	BAKSTP = 0;
	MAGFLG = 0;
	PASFLG = 0;
	NORSCL = 0;
	HLFSCL = 10;
	BAKSCL = 11;
	TXBFLG = 0;
	TXB_U = 0;
	tcaret = 0;
	tlen = 0;
	RLIGHT = 0;
	MLIGHT = 0;
}

// Public Interface
void Viewer::setup_opengl()
{
	glEnable(GL_LINE_SMOOTH);
	glClearColor(bgColor[0], bgColor[1], bgColor[2], 0.0);
	glViewport(0, 0, oslink.width, oslink.height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0, oslink.width, 0, oslink.height);
}

void Viewer::setVid()
{
	if (VDGINV == false)
	{
		// white on black
		bgColor[0] = 0.0;
		bgColor[1] = 0.0;
		bgColor[2] = 0.0;

		fgColor[0] = 1.0;
		fgColor[1] = 1.0;
		fgColor[2] = 1.0;
	}
	else
	{
		// black on white
		bgColor[0] = 1.0;
		bgColor[1] = 1.0;
		bgColor[2] = 1.0;

		fgColor[0] = 0.0;
		fgColor[1] = 0.0;
		fgColor[2] = 0.0;
	}
	glClearColor(bgColor[0], bgColor[1], bgColor[2], 0.0);
}

void Viewer::switchVid()
{
	if (VDGINV == false)
	{
		VDGINV = true;
	}
	else
	{
		VDGINV = false;
	}
	setVid();
}

// This is the main renderer routine.  It draws either
// the map, or the 3D/Examine-Status-Text Area.
void Viewer::draw_game()
{
	if (UPDATE == 0)
	{
		return;
	}
	if (DSPMOD == 0)
	{
		// Draw Map
		glClearColor(0.0, 0.0, 0.0, 0.0);
		glClear(GL_COLOR_BUFFER_BIT);
		glClearColor(bgColor[0], bgColor[1], bgColor[2], 0.0);
		glLoadIdentity();
		MAPPER();
		SDL_GL_SwapBuffers();
	}
	else
	{
		glClear(GL_COLOR_BUFFER_BIT);

		// Draw View Port (3D or Examine or Prepare!)
		glLoadIdentity();
		glColor3fv(fgColor);
		switch (DSPMOD)
		{
		case 1:
			VIEWER();
			break;
		case 2:
			clearArea(&TXTEXA);
			EXAMIN();
			drawArea(&TXTEXA);
			if (player.PTORCH != -1)
			{
				drawTorchHighlite();
			}
			break;
		case 3:
			drawArea(&TXTEXA);
			break;
		default:
			break;
		}

		// Draw Status Line
		drawArea(&TXTSTS);

		// Draw Text Area
		drawArea(&TXTPRI);

		SDL_GL_SwapBuffers();
	}
	UPDATE = 0;
}

// This is the renderer method used to do the wizard
// fade in/out.  It's only used during the opening.  It
// is syncronized with the 30Hz buzz and the wizard
// crashing sound.
bool Viewer::draw_fade()
{	
	delay1 = delay2 = SDL_GetTicks();

	if ((!done && delay1 > delay + buzzStep) && fadeVal != 0)
	{
		glClear(GL_COLOR_BUFFER_BIT);
	}
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	drawArea(&TXTSTS);

	glColor3fv(fgColor);

	if ((!done && delay1 > delay + buzzStep) && fadeVal != 0)
	{
		// Set volume of buzz
		Mix_Volume(fadChannel, ((32 - VCTFAD) / 2) * (MIX_MAX_VOLUME / 16) );

		glLoadIdentity();
		drawVectorList(W1_VLA);
		SDL_GL_SwapBuffers();
		VCTFAD += fadeVal;
		if ((VCTFAD & 0x80) != 0)
		{
			// do sound crash
			Mix_HaltChannel(fadChannel);
			Mix_Volume(fadChannel, MIX_MAX_VOLUME);
			Mix_PlayChannel(fadChannel, creature.kaboom, 0);
			while (Mix_Playing(fadChannel) == 1)
			{
				// Call keyboard routine ???
			}
		
			VCTFAD = 0;
			fadeVal = 0;
		}
		if (VCTFAD == 32 && fadeVal == 2)
		{
			done = true;
			clearArea(&TXTPRI);
			drawArea(&TXTPRI);
		}
		delay = SDL_GetTicks();
	}

	if (VCTFAD == 0 && fadeVal == 0)
	{
		glLoadIdentity();
		drawVectorList(W1_VLA);
		drawArea(&TXTPRI);
		SDL_GL_SwapBuffers();
		delay2 = SDL_GetTicks();
		if (delay2 > delay + midPause)
		{
			// do sound crash
			Mix_PlayChannel(fadChannel, creature.kaboom, 0);
			while (Mix_Playing(fadChannel) == 1)
			{
				// Call keyboard routine ???
			}
		
			fadeVal = 2;
			delay = SDL_GetTicks();
			
			Mix_PlayChannel(fadChannel, creature.buzz, -1);
		}
	}

	if (!done)
	{
		delay1 = delay2 = SDL_GetTicks();
	}
	return done;
}

// Same as above, but used for the intermission
void Viewer::enough_fade()
{	
	delay1 = delay2 = SDL_GetTicks();

	if ((!done && delay1 > delay + buzzStep) && fadeVal != 0)
	{
		glClear(GL_COLOR_BUFFER_BIT);
	}
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	drawArea(&TXTSTS);

	glColor3fv(fgColor);

	if ((!done && delay1 > delay + buzzStep) && fadeVal != 0)
	{
		// Set volume of buzz
		Mix_Volume(fadChannel, ((32 - VCTFAD) / 2) * (MIX_MAX_VOLUME / 16) );

		glLoadIdentity();
		drawVectorList(W1_VLA);
		SDL_GL_SwapBuffers();
		VCTFAD += fadeVal;
		if ((VCTFAD & 0x80) != 0)
		{
			displayEnough();
			drawArea(&TXTPRI);
			SDL_GL_SwapBuffers();

			// do sound crash
			Mix_HaltChannel(fadChannel);
			Mix_Volume(fadChannel, MIX_MAX_VOLUME);
			Mix_PlayChannel(fadChannel, creature.kaboom, 0);
			while (Mix_Playing(fadChannel) == 1)
			{
				// Call keyboard routine ???
			}
		
			VCTFAD = 0;
			fadeVal = 0;
		}
		if (VCTFAD == 32 && fadeVal == 2)
		{
			done = true;
			clearArea(&TXTPRI);
			drawArea(&TXTPRI);
		}
		delay = SDL_GetTicks();
	}

	if (VCTFAD == 0 && fadeVal == 0)
	{
		glLoadIdentity();
		drawVectorList(W1_VLA);
		VCTFAD += fadeVal;
		drawArea(&TXTPRI);
		SDL_GL_SwapBuffers();
		delay2 = SDL_GetTicks();
		if (delay2 > delay + midPause)
		{
			// do sound crash
			Mix_PlayChannel(fadChannel, creature.kaboom, 0);
			while (Mix_Playing(fadChannel) == 1)
			{
				// Call keyboard routine ???
			}
		
			fadeVal = 2;
			delay = SDL_GetTicks();
			
			Mix_PlayChannel(fadChannel, creature.buzz, -1);
		}
	}

	if (!done)
	{
		delay1 = delay2 = SDL_GetTicks();
	}
}

// Same as above, but used for death & victory
void Viewer::death_fade(int WIZ[])
{	
	delay1 = SDL_GetTicks();

	if ((delay1 > delay + buzzStep) && fadeVal != 0)
	{
		// Set volume of buzz
		Mix_Volume(fadChannel, ((32 - VCTFAD) / 2) * (MIX_MAX_VOLUME / 16) );

		glClear(GL_COLOR_BUFFER_BIT);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		drawArea(&TXTSTS);
		glColor3fv(fgColor);
		glLoadIdentity();
		drawVectorList(WIZ);
		SDL_GL_SwapBuffers();
		VCTFAD += fadeVal;
		if ((VCTFAD & 0x80) != 0)
		{
			// do sound crash
			Mix_HaltChannel(fadChannel);
			Mix_Volume(fadChannel, MIX_MAX_VOLUME);
			glLoadIdentity();
			drawArea(&TXTPRI);
			SDL_GL_SwapBuffers();
			Mix_PlayChannel(fadChannel, creature.kaboom, 0);
			while (Mix_Playing(fadChannel) == 1)
			{
				// Call keyboard routine ???
			}
		
			VCTFAD = 0;
			fadeVal = 0;
			done = true;
		}
		delay = SDL_GetTicks();
	}

	if (fadeVal == 0)
	{
		glLoadIdentity();
		drawVectorList(WIZ);
		drawArea(&TXTPRI);
		SDL_GL_SwapBuffers();
	}
}

void Viewer::displayCopyright()
{
	TXB_U = &TXTSTS;
	--TXBFLG;
	OUTSTI(copyright);
	TXBFLG = 0;
}

void Viewer::displayWelcomeMessage()
{
	OUTSTI(welcome1);
	OUTSTI(welcome2);
}

void Viewer::displayDeath()
{
	OUTSTI(death);
}

void Viewer::displayWinner()
{
	OUTSTI(winner1);
	OUTSTI(winner2);
}

void Viewer::displayEnough()
{
	OUTSTI(enough1);
	OUTSTI(enough2);
}

void Viewer::displayPrepare()
{
	clearArea(&TXTEXA);
	TXB_U = &TXTEXA;
	TXB_U->caret = (32*9)+12;
	--TXBFLG;
	OUTSTI(prepstr);
	TXBFLG = 0;
}

void Viewer::drawTorchHighlite()
{
	int x1, y1, x2, y2;
	x1 = tcaret - ((tcaret / 32) * 32);
	y1 = tcaret / 32;
	x2 = x1 + tlen;
	y2 = y1 + 1;
	glLoadIdentity();
	glColor3fv(fgColor);
	glBegin(GL_QUADS);
	glVertex2f(crd.newX(x1*8), crd.newY(y1*8));
	glVertex2f(crd.newX(x2*8), crd.newY(y1*8));
	glVertex2f(crd.newX(x2*8), crd.newY(y2*8));
	glVertex2f(crd.newX(x1*8), crd.newY(y2*8));
	glEnd();
	glColor3fv(bgColor);
	object.OBJNAM(player.PTORCH);
	drawString_internal(x1, y1, parser.TOKEN, tlen);
}

void Viewer::drawArea(TXB * a)
{
	int cnt = 0;

	if (a->top == 19)
	{
		glLoadIdentity();
		glColor3fv(fgColor);
		glBegin(GL_QUADS);
		glVertex2f(crd.newX(0*8), crd.newY(19*8));
		glVertex2f(crd.newX(32*8), crd.newY(19*8));
		glVertex2f(crd.newX(32*8), crd.newY((20*8)));
		glVertex2f(crd.newX(0*8), crd.newY((20*8)));
		glEnd();
		glColor3fv(bgColor);
	}
	else
	{
		glLoadIdentity();
		glColor3fv(bgColor);
		glBegin(GL_QUADS);
		glVertex2f(crd.newX(0*8), crd.newY(20*8));
		glVertex2f(crd.newX(33*8), crd.newY(20*8));
		glVertex2f(crd.newX(33*8), crd.newY((24*8)));
		glVertex2f(crd.newX(0*8), crd.newY((24*8)));
		glEnd();
		glColor3fv(fgColor);
	}

	while (cnt < a->len)
	{
		drawString(0, a->top + cnt / 32, a->area + cnt, 32);
		cnt += 32;
	}
}

void Viewer::clearArea(TXB * a)
{
	strcpy(a->area, "                                ");
	int cnt = 32;
	while (cnt < a->len)
	{
		strcat(a->area, "                                ");
		cnt += 32;
	}
	a->caret = 0;
}

// This method checks if the screen needs updated.
// Called from the scheduler every 3 tenths of a second.
int Viewer::LUKNEW()
{
	// Update Task's next_time
	scheduler.TCBLND[scheduler.luknewTCB].next_time = scheduler.curTime +
		scheduler.TCBLND[scheduler.luknewTCB].frequency;

	if (DSPMOD != 0)
	{
		return 0;
	}

	NEWLUK = 0;
	PUPDAT();
	return 0;
}

// Updates the screen.
void Viewer::PUPDAT()
{
	if (player.FAINT != 0)
	{
		return;
	}

	PUPSUB();

	--UPDATE;
	draw_game();
}

// Sets lighting values.
void Viewer::PUPSUB()
{
	dodBYTE A, B;
	A = player.PRLITE;
	B = player.PMLITE;

	if (player.PTORCH != -1)
	{
		A += object.OCBLND[player.PTORCH].P_OCXX1;
		B += object.OCBLND[player.PTORCH].P_OCXX2;
	}
	RLIGHT = A;
	MLIGHT = B;
}

// Updates the Left and Right hand in the status line
void Viewer::STATUS()
{
	int	ctr, len, offset, idx;
	for (ctr = 0; ctr < 15; ++ctr)
	{
		statArea[ctr] = ' ';
		statArea[ctr + 17] = ' ';
	}
	idx = ((player.PLHAND & 0x8000) != 0) ? -1 : player.PLHAND;
	object.OBJNAM(idx);
	ctr = 0;
	while (parser.TOKEN[ctr] != 0xFF)
	{
		if (parser.TOKEN[ctr] == 0)
		{
			statArea[ctr] = ' ';
		}
		else
		{
			statArea[ctr] = parser.TOKEN[ctr] | 64;
		}
		++ctr;
	}
	idx = ((player.PRHAND & 0x8000) != 0) ? -1 : player.PRHAND;
	object.OBJNAM(idx);
	
	ctr = 0;
	while (parser.TOKEN[ctr] != 0xFF)
	{
		++ctr;
	}

	len = ctr;
	ctr = 32 - len;
	offset = ctr;
	while (ctr < 32)
	{
		if (parser.TOKEN[ctr - offset] == 0)
		{
			statArea[ctr] = ' ';
		}
		else
		{
			statArea[ctr] = parser.TOKEN[ctr - offset] | 64;
		}
		++ctr;
	}
}

void Viewer::OUTSTI(dodBYTE * comp)
{
	int c;
	parser.EXPAND(comp, &c, 0);
	OUTSTR(&parser.STRING[1]);
}

void Viewer::OUTSTR(dodBYTE * str)
{
	int ctr = 0;
	while (*(str + ctr) != 0xFF)
	{
		OUTCHR(*(str + ctr));
		++ctr;
	}
}

void Viewer::OUTCHR(dodBYTE c)
{
	if (TXBFLG == 0)
	{
		TXB_U = &TXTPRI;
	}

	TXTXXX(c);
	if (TXB_U->caret == TXB_U->len && TXB_U->top != 19)
	{
		TXTSCR();
	}
	--UPDATE;
}

void Viewer::TXTXXX(dodBYTE c)
{
	if (c == parser.I_BS)
	{
		// backspace
		if (TXB_U->caret > 0)
		{
			TXB_U->caret -= 1;
		}
		return;
	}
	if (c == parser.I_CR)
	{
		// carriage return
		TXB_U->caret += 32;
		TXB_U->caret = 32 * (TXB_U->caret / 32);
		return;
	}

	if (c == parser.I_SP)
	{
		TXB_U->area[TXB_U->caret] = ' ';
	}
	else if (c >= 1 && c <= 26)
	{
		TXB_U->area[TXB_U->caret] = (c | 64);
	}
	else if (c == parser.I_EXCL)
	{
		TXB_U->area[TXB_U->caret] = '!';
	}
	else if (c == parser.I_BAR)
	{
		TXB_U->area[TXB_U->caret] = '_';
	}
	else if (c == parser.I_QUES)
	{
		TXB_U->area[TXB_U->caret] = '?';
	}
	else if (c == parser.I_DOT)
	{
		TXB_U->area[TXB_U->caret] = '.';
	}
	else if (c == parser.I_SHL)
	{
		TXB_U->area[TXB_U->caret] = '<';
	}
	else if (c == parser.I_SHR)
	{
		TXB_U->area[TXB_U->caret] = '>';
	}
	else if (c == parser.I_LHL)
	{
		TXB_U->area[TXB_U->caret] = '{';
	}
	else if (c == parser.I_LHR)
	{
		TXB_U->area[TXB_U->caret] = '}';
	}

	++TXB_U->caret;
}

void Viewer::TXTSCR()
{
	int ctr;
	for (ctr = 0; ctr < TXB_U->len - 32; ++ctr)
	{
		TXB_U->area[ctr] = TXB_U->area[ctr + 32];
	}
	for (ctr = TXB_U->len - 32; ctr < TXB_U->len; ++ctr)
	{
		TXB_U->area[ctr] = 0;
	}
	TXB_U->caret = TXB_U->len - 32;
	if (player.PTORCH != -1 && TXB_U->len > 128)
	{
		tcaret -= 32;
	}
}

void Viewer::PROMPT()
{
	OUTSTR(parser.M_PROM1);
}

// This is the 3D-Viewport rendering routine
void Viewer::VIEWER()
{
	dodBYTE a, b, x, u, ftctr, vft;
	int creNum, objIdx;

	dungeon.RANGE = 0;
	dungeon.DROW.setRC(player.PROW, player.PCOL);

	do
	{
		SETSCL();
		a = dungeon.MAZLND[dungeon.RC2IDX(dungeon.DROW.row, dungeon.DROW.col)];
		u = 0;
		x = 4;
		do
		{
			b = a;
			b = (b & 3);
			dungeon.NEIBOR[u+4] = b;
			dungeon.NEIBOR[u] = b;
			++u;
			a >>= 2;
			--x;
		} while (x != 0);

		b = player.PDIR;
		u = b;

		for (ftctr = 0; ftctr < 3; ++ftctr)
		{
			b = dungeon.NEIBOR[u + FLATAB[ftctr]];
			if (b == dungeon.HF_SDR)
			{
				--MAGFLG;
				DRAWIT(FLATABv[ftctr][b]);
				b = dungeon.HF_WAL;
			}
			DRAWIT(FLATABv[ftctr][b]);
		}

		creNum = creature.CFIND2(dungeon.DROW);
		if (creNum != -1)
		{
			CMRDRW(FWDCRE[creature.CCBLND[creNum].P_CCTYP], creNum);
		}

		PDRAW(LPK_VLA, 3, u);
		PDRAW(RPK_VLA, 1, u);

		// Draw vertical features
		vft = dungeon.VFIND(dungeon.DROW);
		if (vft == 0xFF)
		{
			DRAWIT(CEI_VLA);
		}
		else
		{
			switch(vft)
			{
			case 0:
				DRAWIT(HUP_VLA);
				break;
			case 1:
				DRAWIT(LAD_VLA);
				DRAWIT(HUP_VLA);
				break;
			case 2:
				DRAWIT(HDN_VLA);
				DRAWIT(CEI_VLA);
				break;
			case 3:
				DRAWIT(LAD_VLA);
				DRAWIT(HDN_VLA);
				DRAWIT(CEI_VLA);
				break;
			default:
				break; // should never get here
			}
		}

		// Draw Objects
		object.OFINDF = 0;
		do
		{
			objIdx = object.OFIND(dungeon.DROW);
			if (objIdx == -1)
				break;
			--MAGFLG;
			DRAWIT(FWDOBJ[object.OCBLND[objIdx].P_OCCLS]);
			DRAWIT(FWDOBJ[object.OCBLND[objIdx].P_OCCLS]);
		} while (true);


		if (dungeon.NEIBOR[u] != 0)
		{
			break;
		}
		dungeon.DROW.row += dungeon.STPTAB[player.PDIR * 2];
		dungeon.DROW.col += dungeon.STPTAB[player.PDIR * 2 + 1];
		++dungeon.RANGE;
	} while (dungeon.RANGE <= 9);
}

// Calculates fade (line-pixelation) based on lighting
void Viewer::SETFAD()
{
	dodBYTE a, b;
	a = RLIGHT;
	if (MAGFLG != 0)
	{
		a = MLIGHT;
		MAGFLG = 0;
	}
	b = 0;
	a -= 7;
	a -= dungeon.RANGE;
	if ((a & 128) != 0)
	{
		--b;
		if ((((a & 128) != 0) && a > 0xF9) || ((a & 128) == 0)) //	if (a > -7)
		{
			b = (1 << (-1 - a));
		}
	}
	VCTFAD = b;
}

// Used by 3D-Viewer, draws a vector list
void Viewer::DRAWIT(int * vl)
{
	SETFAD();
	drawVectorList(vl);
}

// Used by 3D-Viewer, checks for around-the-corner creature
void Viewer:: PDRAW(int * vl, dodBYTE dir, dodBYTE pdir)
{
	RowCol side;
	dodBYTE DIR;
	int creNum;
	if (dungeon.NEIBOR[pdir + dir] != 0)
		return;
	DIR = ((dir + player.PDIR) & 3);
	side.row = dungeon.DROW.row + dungeon.STPTAB[DIR * 2];
	side.col = dungeon.DROW.col + dungeon.STPTAB[DIR * 2 + 1];
	creNum = creature.CFIND2(side);
	if (creNum == -1)
		return;
	CMRDRW(vl, creNum);
}

// Prepares for drawing creature with either magical or physical lighting
void Viewer::CMRDRW(int * vl, int creNum)
{
	if (creature.CCBLND[creNum].P_CCMGO != 0)
		--MAGFLG;
	DRAWIT(vl);
}

// Sets the perspective scale
void Viewer::SETSCL()
{
	int idx = HLFSCL;
	if (HLFSTP == 0)
	{
		++idx;
		if (BAKSTP == 0)
		{
			idx = 0;
		}
	}
	VXSCAL = Scale[idx + dungeon.RANGE];
	VYSCAL = Scale[idx + dungeon.RANGE];
}

void Viewer::EXAMIN()
{
	TXB_U = &TXTEXA;
	--TXBFLG;
	NEWLIN = 0;
	TXB_U->caret = 10;
	OUTSTI(exam1);

	// check for creature
	if (creature.CFIND2(RowCol(player.PROW, player.PCOL)) != -1)
	{
		TXB_U->caret += 11;
		OUTSTI(exam2);
	}

	// check for objects on floor
	int idx;
	object.OFINDF = 0;
	do
	{
		idx = object.OFIND(RowCol(player.PROW, player.PCOL));
		if (idx != -1)
		{
			PRTOBJ(idx, false);
		}
	} while (idx != -1);

	if (NEWLIN != 0)
	{
		PCRLF();
	}

	int ctr = 32;
	do
	{
		OUTCHR(parser.I_EXCL);
		--ctr;
	} while (ctr != 0);

	TXB_U->caret += 12;
	OUTSTI(exam3);

	// bag contents
	ctr = player.BAGPTR;
	while (ctr != -1)
	{
		// check for torch
		if (ctr == player.PTORCH)
		{
			PRTOBJ(ctr, true);
		}
		else
		{
			PRTOBJ(ctr, false);
		}
		ctr = object.OCBLND[ctr].P_OCPTR;
	}

	TXBFLG = 0;
}

void Viewer::PCRLF()
{
	OUTCHR(parser.I_CR);
	NEWLIN = 0;
}

void Viewer::PRTOBJ(int X, bool highlite)
{
	object.OBJNAM(X);
	if (highlite)
	{
		tcaret = TXB_U->caret;
		int ctr = 0;
		while (parser.TOKEN[ctr] != 0xFF)
		{
			++ctr;
		}
		tlen = ctr;
	}
	OUTSTR(parser.TOKEN);
	NEWLIN = (~NEWLIN);
	if (NEWLIN != 0)
	{
		TXB_U->caret += 16;
		TXB_U->caret = (TXB_U->caret / 16) * 16;
	}
	else
	{
		PCRLF();
	}
}

// Draws the map; MAPFLG bool determines VISION or SEER mode
void Viewer::MAPPER()
{
	int mazIdx, objIdx, creIdx, vftIdx;
	RowCol rc;
	dodBYTE a;
	bool vftOnce;
	
	dungeon.DROW.row = 31;
	dungeon.DROW.col = 31;
	glColor3f(1.0,1.0,1.0);
	do
	{
		mazIdx = dungeon.RC2IDX(dungeon.DROW.row, dungeon.DROW.col);
		if (dungeon.MAZLND[mazIdx] == 0xFF)
		{
			glBegin(GL_QUADS);
			glVertex2f(crd.newX(dungeon.DROW.col * 8), crd.newY(dungeon.DROW.row * 6));
			glVertex2f(crd.newX(dungeon.DROW.col * 8), crd.newY((dungeon.DROW.row + 1) * 6));
			glVertex2f(crd.newX((dungeon.DROW.col + 1) * 8), crd.newY((dungeon.DROW.row + 1) * 6));
			glVertex2f(crd.newX((dungeon.DROW.col + 1) * 8), crd.newY(dungeon.DROW.row * 6));
			glEnd();
		}
		--dungeon.DROW.col;
		if (dungeon.DROW.col == 0xFF)
		{
			--dungeon.DROW.row;
			dungeon.DROW.col = 31;
		}
	} while (dungeon.DROW.row != 0xFF);

	if (MAPFLG == true)
	{
		// Mark Objects
		object.OFINDF = 0;
		do
		{
			objIdx = object.FNDOBJ();
			if (objIdx == -1)
				break;
			if (object.OCBLND[objIdx].P_OCOWN != 0)
				continue;
			rc.row = object.OCBLND[objIdx].P_OCROW;
			rc.col = object.OCBLND[objIdx].P_OCCOL;
			glBegin(GL_QUADS);
			glVertex2f(crd.newX((rc.col * 8) + 4), crd.newY((rc.row * 6) + 2));
			glVertex2f(crd.newX((rc.col * 8) + 4), crd.newY((rc.row * 6) + 4));
			glVertex2f(crd.newX((rc.col * 8) + 5), crd.newY((rc.row * 6) + 4));
			glVertex2f(crd.newX((rc.col * 8) + 5), crd.newY((rc.row * 6) + 2));
			glEnd();
		} while (true);

		// Mark Creatures
		creIdx = -1;
		do
		{
			++creIdx;
			if (creIdx == 32)
				break;
			if (creature.CCBLND[creIdx].P_CCUSE == 0)
				continue;
			rc.row = creature.CCBLND[creIdx].P_CCROW;
			rc.col = creature.CCBLND[creIdx].P_CCCOL;
			glBegin(GL_QUADS);
			glVertex2f(crd.newX((rc.col * 8) + 1), crd.newY((rc.row * 6) + 2));
			glVertex2f(crd.newX((rc.col * 8) + 1), crd.newY((rc.row * 6) + 4));
			glVertex2f(crd.newX((rc.col * 8) + 2), crd.newY((rc.row * 6) + 4));
			glVertex2f(crd.newX((rc.col * 8) + 2), crd.newY((rc.row * 6) + 2));
			
			glVertex2f(crd.newX((rc.col * 8) + 5), crd.newY((rc.row * 6) + 2));
			glVertex2f(crd.newX((rc.col * 8) + 5), crd.newY((rc.row * 6) + 4));
			glVertex2f(crd.newX((rc.col * 8) + 6), crd.newY((rc.row * 6) + 4));
			glVertex2f(crd.newX((rc.col * 8) + 6), crd.newY((rc.row * 6) + 2));
			
			glVertex2f(crd.newX((rc.col * 8) + 3), crd.newY((rc.row * 6) + 1));
			glVertex2f(crd.newX((rc.col * 8) + 3), crd.newY((rc.row * 6) + 5));
			glVertex2f(crd.newX((rc.col * 8) + 4), crd.newY((rc.row * 6) + 5));
			glVertex2f(crd.newX((rc.col * 8) + 4), crd.newY((rc.row * 6) + 1));
			glEnd();
		} while (true);
	}

	// Mark Player
	rc.row = player.PROW;
	rc.col = player.PCOL;
	glBegin(GL_QUADS);
	glVertex2f(crd.newX((rc.col * 8) + 2), crd.newY((rc.row * 6) + 1));
	glVertex2f(crd.newX((rc.col * 8) + 2), crd.newY((rc.row * 6) + 2));
	glVertex2f(crd.newX((rc.col * 8) + 3), crd.newY((rc.row * 6) + 2));
	glVertex2f(crd.newX((rc.col * 8) + 3), crd.newY((rc.row * 6) + 1));

	glVertex2f(crd.newX((rc.col * 8) + 5), crd.newY((rc.row * 6) + 1));
	glVertex2f(crd.newX((rc.col * 8) + 5), crd.newY((rc.row * 6) + 2));
	glVertex2f(crd.newX((rc.col * 8) + 6), crd.newY((rc.row * 6) + 2));
	glVertex2f(crd.newX((rc.col * 8) + 6), crd.newY((rc.row * 6) + 1));

	glVertex2f(crd.newX((rc.col * 8) + 3), crd.newY((rc.row * 6) + 2));
	glVertex2f(crd.newX((rc.col * 8) + 3), crd.newY((rc.row * 6) + 4));
	glVertex2f(crd.newX((rc.col * 8) + 5), crd.newY((rc.row * 6) + 4));
	glVertex2f(crd.newX((rc.col * 8) + 5), crd.newY((rc.row * 6) + 2));

	glVertex2f(crd.newX((rc.col * 8) + 2), crd.newY((rc.row * 6) + 4));
	glVertex2f(crd.newX((rc.col * 8) + 2), crd.newY((rc.row * 6) + 5));
	glVertex2f(crd.newX((rc.col * 8) + 3), crd.newY((rc.row * 6) + 5));
	glVertex2f(crd.newX((rc.col * 8) + 3), crd.newY((rc.row * 6) + 4));

	glVertex2f(crd.newX((rc.col * 8) + 5), crd.newY((rc.row * 6) + 4));
	glVertex2f(crd.newX((rc.col * 8) + 5), crd.newY((rc.row * 6) + 5));
	glVertex2f(crd.newX((rc.col * 8) + 6), crd.newY((rc.row * 6) + 5));
	glVertex2f(crd.newX((rc.col * 8) + 6), crd.newY((rc.row * 6) + 4));
	glEnd();

	// Mark Vertical Features
	vftIdx = dungeon.VFTPTR;
	vftOnce == false;
	do
	{
		a = dungeon.VFTTAB[vftIdx++];
		if (a == 0xFF)
		{
			if (vftOnce == false)
			{
				vftOnce = true;
				continue;
			}
			else
			{
				break;
			}
		}
		rc.row = dungeon.VFTTAB[vftIdx++];
		rc.col = dungeon.VFTTAB[vftIdx++];

		glBegin(GL_QUADS);
		glVertex2f(crd.newX((rc.col * 8) + 2), crd.newY((rc.row * 6) + 1));
		glVertex2f(crd.newX((rc.col * 8) + 2), crd.newY((rc.row * 6) + 5));
		glVertex2f(crd.newX((rc.col * 8) + 3), crd.newY((rc.row * 6) + 5));
		glVertex2f(crd.newX((rc.col * 8) + 3), crd.newY((rc.row * 6) + 1));

		glVertex2f(crd.newX((rc.col * 8) + 5), crd.newY((rc.row * 6) + 1));
		glVertex2f(crd.newX((rc.col * 8) + 5), crd.newY((rc.row * 6) + 5));
		glVertex2f(crd.newX((rc.col * 8) + 6), crd.newY((rc.row * 6) + 5));
		glVertex2f(crd.newX((rc.col * 8) + 6), crd.newY((rc.row * 6) + 1));

		glVertex2f(crd.newX((rc.col * 8) + 3), crd.newY((rc.row * 6) + 1));
		glVertex2f(crd.newX((rc.col * 8) + 3), crd.newY((rc.row * 6) + 2));
		glVertex2f(crd.newX((rc.col * 8) + 5), crd.newY((rc.row * 6) + 2));
		glVertex2f(crd.newX((rc.col * 8) + 5), crd.newY((rc.row * 6) + 1));

		glVertex2f(crd.newX((rc.col * 8) + 3), crd.newY((rc.row * 6) + 4));
		glVertex2f(crd.newX((rc.col * 8) + 3), crd.newY((rc.row * 6) + 5));
		glVertex2f(crd.newX((rc.col * 8) + 5), crd.newY((rc.row * 6) + 5));
		glVertex2f(crd.newX((rc.col * 8) + 5), crd.newY((rc.row * 6) + 4));
		glEnd();

	} while (true);
}

// Draws non-font vector lists
void Viewer::drawVectorList(int VLA[])
{
	int numLists = VLA[0];
	int curList = 0;
	int numVertices;
	int curVertex;
	int	ctr = 1;
	dodSHORT x0, y0, x1, y1;

	if (VCTFAD == 0xFF)
	{
		return;
	}

	while (curList < numLists)
	{
		numVertices = VLA[ctr];
		++ctr;
		curVertex = 0;
		while (curVertex < (numVertices - 1) )
		{
			x0 = ScaleX(VLA[ctr]) + VCNTRX;
			y0 = ScaleY(VLA[ctr+1]) + VCNTRY;
			x1 = ScaleX(VLA[ctr+2]) + VCNTRX;
			y1 = ScaleY(VLA[ctr+3]) + VCNTRY;
			drawVector((signed short)x0, (signed short)y0, (signed short)x1, (signed short)y1);
			ctr += 2;
			++curVertex;
		}
		++curList;
		ctr += 2;
	}
}

// Scales X-coordinate
dodSHORT Viewer::ScaleX(int x)
{
	dodBYTE a, b;
	dodSHORT d;
	b = (dodBYTE) x;
	a = VXSCAL;
	b -= (VCNTRX);
	if ((b & 0x80) != 0)
	{
		b = (~b) + 1;
		d = a * b;
		d = (~d) + 1;
		d = (d >> 7) | 0xFE00;
	}
	else
	{
		d = a * b;
		d = (d >> 7);
	}
	return d;
}

// Scales Y-coordinate
dodSHORT Viewer::ScaleY(int y)
{
	dodBYTE a, b;
	dodSHORT d;
	b = (dodBYTE) y;
	a = VYSCAL;
	b -= (VCNTRY);
	if ((b & 0x80) != 0)
	{
		b = (~b) + 1;
		d = a * b;
		d = (~d) + 1;
		d = (d >> 7) | 0xFE00;
	}
	else
	{
		d = a * b;
		d = (d >> 7);
	}
	return d;
}

// Draws font vectors
void Viewer::drawVectorListAQ(int VLA[])
{
	int numQuads = VLA[0];
	int curQuad = 0;
	int	ctr = 1;

	glBegin(GL_QUADS);
	while (curQuad < numQuads)
	{
		glVertex2f(crd.newXa(((double) VLA[ctr]) / 4.0L),crd.newYa(((double) VLA[ctr+1]) / 4.0L));
		glVertex2f(crd.newXa(((double) VLA[ctr+2]) / 4.0L),crd.newYa(((double) VLA[ctr+3]) / 4.0L));
		glVertex2f(crd.newXa(((double) VLA[ctr+4]) / 4.0L),crd.newYa(((double) VLA[ctr+5]) / 4.0L));
		glVertex2f(crd.newXa(((double) VLA[ctr+6]) / 4.0L),crd.newYa(((double) VLA[ctr+7]) / 4.0L));
		ctr += 8;
		++curQuad;
	}
	glEnd();
}

// Draws a character
void Viewer::drawCharacter(char c)
{
	if (c >= 'A' && c <= 'Z')
	{
		drawVectorListAQ(AZ_VLA[c-64]);
	}
	switch (c)
	{
	case ('<'):
		drawVectorListAQ(AZ_VLA[31]);
		break;
	case ('>'):
		drawVectorListAQ(AZ_VLA[32]);
		break;
	case ('{'):
		drawVectorListAQ(AZ_VLA[33]);
		break;
	case ('}'):
		drawVectorListAQ(AZ_VLA[34]);
		break;
	case (' '):
		break;
	case ('.'):
		drawVectorListAQ(AZ_VLA[27]);
		break;
	case ('_'):
		drawVectorListAQ(AZ_VLA[28]);
		break;
	case ('!'):
		drawVectorListAQ(AZ_VLA[29]);
		break;
	case ('?'):
		drawVectorListAQ(AZ_VLA[30]);
		break;
	default:
		break;
	}
}

void Viewer::drawString_internal(int x, int y, dodBYTE * str, int len)
{
	int ctr;
	char c;
	glLoadIdentity();
	glTranslatef(crd.newX(x*8), crd.newY(((y+1)*8)), 0.0);
	for (ctr=0; ctr < len; ++ctr)
	{
		c = dod_to_ascii(*(str + ctr));
		drawCharacter(c);
		glTranslatef(crd.newXa(8), 0.0, 0.0);
	}
}

char Viewer::dod_to_ascii(dodBYTE c)
{
	if (c >= 1 && c <= 26) return (c | 64);
	if (c == parser.I_SP) return ' ';
	if (c == parser.I_EXCL) return '!';
	if (c == parser.I_BAR) return '_';
	if (c == parser.I_QUES) return '?';
	if (c == parser.I_DOT) return '.';
	if (c == parser.I_SHL) return '<';
	if (c == parser.I_SHR) return '>';
	if (c == parser.I_LHL) return '{';
	if (c == parser.I_LHR) return '}';
	return ' ';
}

// Draws a string
void Viewer::drawString(int x, int y, char * str, int len)
{
	int ctr;
	glLoadIdentity();
	glTranslatef(crd.newX(x*8), crd.newY(((y+1)*8)), 0.0);
	for (ctr=0; ctr < len; ++ctr)
	{
		drawCharacter(*(str + ctr));
		glTranslatef(crd.newXa(8), 0.0, 0.0);
	}
}

// Draws a line
void Viewer::drawVector(int X0, int Y0, int X1, int Y1)
{
	int XL, YL, FADCNT, L;
	double DX, DY, XX, YY;

	if (VCTFAD == 0xFF)
	{
		return;
	}
	FADCNT = VCTFAD + 1;
	XL = (X1 > X0) ? (X1 - X0) : (X0 - X1);
	YL = (Y1 > Y0) ? (Y1 - Y0) : (Y0 - Y1);
	L = (XL > YL) ? XL : YL;
	if (L == 0)
	{
		return;
	}
	DX = ((double) XL / (double) L) * ((X0 < X1) ? 1 : -1);
	DY = ((double) YL / (double) L) * ((Y0 < Y1) ? 1 : -1);
	XX = X0 + 0.5;
	YY = Y0 + 0.5;
	do
	{
		--FADCNT;
		if (FADCNT == 0)
		{
			FADCNT = VCTFAD + 1;
			if ((int) XX >= 0 && (int) XX <= 255 &&
				(int) YY >= 0 && (int) YY <= 151)
			{
				plotPoint((int) XX, (int) YY);
			}
		}
		XX += DX;
		YY += DY;
		--L;
	} while (L > 0);
}

// Draws one pixel
void Viewer::plotPoint(double X, double Y)
{
	glBegin(GL_QUADS);
	glVertex2f(crd.newX(X),		crd.newY(Y));
	glVertex2f(crd.newX(X+1),	crd.newY(Y));
	glVertex2f(crd.newX(X+1),	crd.newY(Y+1));
	glVertex2f(crd.newX(X),		crd.newY(Y+1));
	glEnd();
}
